from typing import Any, Dict, List, AsyncGenerator
from openai import AsyncOpenAI
from openai.types.chat import ChatCompletion, ChatCompletionChunk
from openai import RateLimitError, APIError
from retry import retry
from typing import Union
import time
from pegasus.client.response import StreamEvent, StreamEventType, TextDelta, ReasoningDelta, ToolCall, TokenUsage, parse_tool_call_arguments
from pegasus.config.config import Config
from pegasus.utils.logger import logger


class LLMClient:
    def __init__(self, config: Config) -> None:
        self._client: AsyncOpenAI | None = None
        self._config = config


    def _get_client(self) -> AsyncOpenAI:
        if self._client is None:
            self._client = AsyncOpenAI(api_key=self._config.api_key, 
                                       base_url=self._config.base_url)
        return self._client
    
    async def close(self) -> None:
        if self._client is not None:
            await self._client.close()
            self._client = None

    def _build_tools(self, tools: List[dict[str, Any]] | None) -> List[dict[str, Any]]:
        """
        example tool:
        {
            "type": "function",
            "function": {
                "name": "search_gutenberg_books",
                "description": "Search for books in the Project Gutenberg library",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "search_terms": {
                            "type": "array",
                            "items": {"type": "string"},
                            "description": "List of search terms to find books"
                        }
                    },
                    "required": ["search_terms"]
                }
            }
        }
        """
        return [{
            "type": "function",
            "function": {
                "name": tool["name"],
                "description": tool.get("description", ""),
                "parameters": tool.get("parameters", {
                    "type": "object",
                    "properties": {},
                }),
                
            }
        } for tool in tools]

    @retry(exceptions=RateLimitError, tries=3, delay=1, backoff=2)
    async def chat_completion(self,
                              messages: List[Dict[str, Any]], 
                              tools: List[dict[str, Any]] | None = None,
                              stream: bool = False
                              ) -> AsyncGenerator[StreamEvent, None]:
        """ 
        example message:
        "messages": [
            {"role": "system", "content": "You are a helpful assistant."},
            {"role": "user", "content": "Hello!"}
        ]
        """
        client: AsyncOpenAI = self._get_client()
        kwargs: Dict[str, Any] = {
            'model': self._config.model_name,
            'messages': messages,
            'stream': stream,
        }
        if tools is not None:
            kwargs['tools'] = self._build_tools(tools)
            kwargs['tool_choice'] = 'auto'
        try:
            if stream:
                async for event in self._stream_response(client, kwargs):
                    yield event
            else:
                yield await self._non_stream_response(client, kwargs)
        except RateLimitError:
            yield StreamEvent(
                type=StreamEventType.ERROR,
                error='Rate limit exceeded',
                text_delta=None,
                finish_reason=None,
                usage=None,
            )
            raise RateLimitError
        except APIError:
            yield StreamEvent(
                type=StreamEventType.ERROR,
                error='API error',
                text_delta=None,
                finish_reason=None,
                usage=None,
            )


    async def _stream_response(self, client: AsyncOpenAI, kwargs: Dict[str, Any]) -> AsyncGenerator[StreamEvent, None]:
        t_start = time.perf_counter()
        response: ChatCompletion = await client.chat.completions.create(**kwargs)

        usage = None
        tool_calls: Dict[str, Any] = {}

        async for chunk in response:
            chunk: ChatCompletionChunk = chunk
            if chunk.usage:
                usage = TokenUsage(
                    prompt_tokens=chunk.usage.prompt_tokens,
                    completion_tokens=chunk.usage.completion_tokens,
                    total_tokens=chunk.usage.total_tokens,
                )
            if not chunk.choices:
                # if no choices then just continue and wait for next chunk
                continue
            choice = chunk.choices[0]
            delta = choice.delta
            finish_reason = None
            if choice.finish_reason:
                finish_reason = choice.finish_reason
            text_delta = None
            reasoning_delta = None
            refusal = None
            if delta.refusal:
                refusal = delta.refusal
            if hasattr(delta, "reasoning") and delta.reasoning:
                reasoning_delta = ReasoningDelta(reasoning=delta.reasoning)
            if delta.content:
                text_delta = TextDelta(content=delta.content)
            if text_delta or reasoning_delta or refusal:
                yield StreamEvent(
                    type=StreamEventType.TEXT_DELTA,
                    text_delta=text_delta,
                    reasoning_delta=reasoning_delta,
                    refusal=refusal,
                    error=None,
                    finish_reason=finish_reason,
                    usage=usage,
                )
            if delta.tool_calls:
                for tool_call_delta in delta.tool_calls:
                    idx = tool_call_delta.index
                    if idx not in tool_calls:
                        tool_calls[idx] = {
                            "id": tool_call_delta.id,
                            "name": "",
                            "arguments": "",
                        }
                    if tool_call_delta.function:
                        if tool_call_delta.function.name:
                            tool_calls[idx]["name"] = tool_call_delta.function.name
                        if tool_call_delta.function.arguments:
                            tool_calls[idx]["arguments"] += tool_call_delta.function.arguments
        t_end = time.perf_counter()
        total_time = t_end - t_start if (t_end - t_start) > 0 else None
        if usage and total_time is not None:
            usage = TokenUsage(
                prompt_tokens=usage.prompt_tokens,
                completion_tokens=usage.completion_tokens,
                total_tokens=usage.total_tokens,
                total_time=total_time,
            )
        _tool_calls: List[ToolCall] = []
        for idx, tool_call in tool_calls.items():
            _tool_calls.append(ToolCall(
                call_id=tool_call["id"],
                name=tool_call["name"],
                arguments=parse_tool_call_arguments(tool_call["arguments"]),
            ))
        yield StreamEvent(
            type=StreamEventType.TOOL_CALL_COMPLETE,
            tool_calls=_tool_calls,
        )
        yield StreamEvent(
            type=StreamEventType.MESSAGE_COMPLETE,
            usage=usage,
            finish_reason=finish_reason,
        )
        
    async def _non_stream_response(self, client: AsyncOpenAI, kwargs: Dict[str, Any]) -> Union[StreamEvent, None]:
        t_start = time.perf_counter()
        response: ChatCompletion = await client.chat.completions.create(**kwargs)
        t_end = time.perf_counter()
        completion_tokens = None
        prompt_tokens = None
        total_tokens = None
        if hasattr(response, "usage"):
            usage = response.usage
            completion_tokens = usage.completion_tokens
            prompt_tokens = usage.prompt_tokens
            total_tokens = usage.total_tokens

        # select choice
        choice = response.choices[0]
        # get message of the choice
        message = choice.message
        # get content of the message
        content = message.content
        # get refusal of the message
        refusal = message.refusal
        # get annotations of the message
        annotations = message.annotations
        # get tool calls of the message
        tool_calls = message.tool_calls
        # get reasoning of the message
        reasoning = None
        if hasattr(message, "reasoning"):
            reasoning = message.reasoning

        # get finish reason
        finish_reason = None
        if choice.finish_reason:
            finish_reason = choice.finish_reason

        # create the text delta
        text_delta = None
        if content:
            text_delta = TextDelta(content=content)

        # create the reasoning delta
        if reasoning:
            reasoning_delta = ReasoningDelta(reasoning=reasoning)


        # create the tool calls
        tool_calls: List[ToolCall] = []
        if message.tool_calls:
            for tool_call in message.tool_calls:
                tool_calls.append(ToolCall(
                    call_id=tool_call.id,
                    name=tool_call.function.name,
                    arguments=parse_tool_call_arguments(tool_call.function.arguments),
                ))

        total_time = (t_end - t_start) if (t_end - t_start) > 0 else None
        usage = TokenUsage(
            prompt_tokens=prompt_tokens or 0,
            completion_tokens=completion_tokens or 0,
            total_tokens=total_tokens or 0,
            total_time=total_time,
        )

        # return the streaming event
        return StreamEvent(
            type=StreamEventType.MESSAGE_COMPLETE,
            text_delta=text_delta,
            reasoning_delta=reasoning_delta,
            usage=usage,
            finish_reason=finish_reason,
            refusal=refusal,
            annotations=annotations,
            error=None,
            tool_calls=tool_calls,
        )


from dataclasses import dataclass, field
from pathlib import Path
from typing import Any
from enum import Enum


class ToolType(str, Enum):
    READ = "read"
    WRITE = "write"
    BASH = "bash"
    NETWORK = "network"
    MEMORY = "memory"
    MCP = "mcp"
    SUB_AGENT = "sub_agent"

@dataclass
class FileDiff:
    path: Path
    old_content: str
    new_content: str
    is_new_file: bool
    is_deletion: bool

    def create_diff(self) -> str:
        import difflib
        old_lines = []
        if self.old_content:
            old_lines = self.old_content.splitlines(keepends=True)
        new_lines = self.new_content.splitlines(keepends=True)
        if old_lines and not old_lines[-1].endswith("\n"):
            old_lines[-1] += "\n"
        if new_lines and not new_lines[-1].endswith("\n"):
            new_lines[-1] += "\n"

        old_name = "/dev/null" if self.is_new_file else self.path.as_posix()
        new_name = "/dev/null" if self.is_deletion else self.path.as_posix()
        deltas = difflib.unified_diff(old_lines, new_lines, fromfile=old_name, tofile=new_name)
        return "\n".join(deltas)


@dataclass
class ToolInvocation:
    cwd: Path
    params: dict[str, Any]

@dataclass
class ToolImage:
    data_url: str
    mime_type: str
    path: str | None = None

    def to_content_part(self) -> dict[str, Any]:
        return {"type": "image_url", "image_url": {"url": self.data_url}}

    def to_metadata(self) -> dict[str, Any]:
        result = {"mime_type": self.mime_type}
        if self.path:
            result["path"] = self.path
        return result


@dataclass
class ToolResult:
    success: bool
    output: str
    exit_code: int | None = None
    error: str | None = None
    metadata: dict[str, Any] = field(default_factory=dict)
    truncated: bool = False
    diff: FileDiff | None = None
    images: list[ToolImage] = field(default_factory=list)

    @classmethod
    def error_result(
        cls,
        error: str,
        output: str = "",
        metadata: dict[str, Any] | None = None,
        images: list[ToolImage] | None = None,
    ) -> 'ToolResult':
        return cls(
            success=False,
            output=output,
            error=error,
            metadata=metadata or {},
            images=images or [],
        )
    
    @classmethod
    def success_result(cls, output: str, **kwargs) -> 'ToolResult':
        return cls(success=True, output=output, **kwargs)

    def to_model_output(self) -> str:
        if self.success:
            return self.output
        return f"Error: {self.error}\n\nOutput: {self.output}"


from pegasus.tools.base import Tool
from pegasus.tools.data import ToolType, ToolInvocation, ToolResult
from pegasus.config.config import Config
from pydantic import BaseModel, Field   
from ddgs import DDGS
from pegasus.utils.logger import logger

class WebSearchParams(BaseModel):
    query: str = Field(..., description="The query to search for")
    max_results: int = Field(10, ge=1, le=100, description="The maximum number of results to return (default 10, max 100, min 1)")


class WebSearchTool(Tool):
    name: str = "websearch"
    description: str = "Search the web for information. Returns search results with titles, urls and snippets"
    type: ToolType = ToolType.NETWORK
    schema: WebSearchParams = WebSearchParams

    def __init__(self, config: Config) -> None:
        self._config = config

    async def _execute(self, invocation: ToolInvocation) -> ToolResult:
        params = WebSearchParams(**invocation.params)
        results = []
        try:
            results = DDGS().text(params.query, region="us-en", safesearch="off", timelimit="y", page=1, backend="auto")
        except Exception as e:
            return ToolResult.error_result(f"Error searching the web with error: {e}")
        if not results:
            return ToolResult.error_result(f"No results found for query: {params.query}")
        
        output_lines = [f'Search results for "{params.query}":']
        for idx, result in enumerate(results[:params.max_results]):
            output_lines.append(f'{idx+1}. Title: {result.get("title")}')
            output_lines.append(f'         URL: {result.get("href")}')
            if result.get("body"):
                output_lines.append(f'         Snippet: {result.get("body")}')
            else:
                output_lines.append(f'         No snippet available')
            output_lines.append('-' * 100)
            output_lines.append('')
        truncated = False
        if len(results) > params.max_results:
            output_lines.append(f'...truncated to {params.max_results} results...')
            truncated = True
        return ToolResult.success_result(output="\n".join(output_lines), truncated=truncated, metadata=dict(results=results[:params.max_results], total_results=len(results)))
    
    async def execute(self, invocation: ToolInvocation) -> ToolResult:
        try:
            return await self._execute(invocation)
        except Exception as e:
            return ToolResult.error_result(f"Error searching the web with error: {e}")


import logging

# ansi color codes for level-based coloring
_GREEN = "\033[32m"
_YELLOW = "\033[33m"
_RED = "\033[31m"
_RESET = "\033[0m"

LEVEL_COLORS = {
    logging.DEBUG: _GREEN,
    logging.INFO: _GREEN,
    logging.WARNING: _YELLOW,
    logging.ERROR: _RED,
    logging.CRITICAL: _RED,
}


class ColoredFormatter(logging.Formatter):
    """Format log records with level-based colors."""

    def format(self, record: logging.LogRecord) -> str:
        color = LEVEL_COLORS.get(record.levelno, _RESET)
        return f"{color}{super().format(record)}{_RESET}"


_handler = logging.StreamHandler()
_handler.setFormatter(
    ColoredFormatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
)
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.addHandler(_handler)
logger.propagate = False


from typing import List
from pegasus.tools.base import Tool
from pegasus.tools.builtin.readfile import ReadFileTool
from pegasus.tools.builtin.writefile import WriteFileTool
from pegasus.tools.builtin.editfile import EditFileTool 
from pegasus.tools.builtin.shell import ShellTool
from pegasus.tools.builtin.websearch import WebSearchTool
from pegasus.tools.builtin.todos import TodosTool
from pegasus.tools.builtin.memory import MemoryTool 
from pegasus.tools.builtin.browser_use import BrowserUseTool
from pegasus.tools.builtin.vision_capture import VisionCaptureTool
from pegasus.tools.builtin.mcp_executor import MCPExecutorTool
from pegasus.tools.builtin.subagent import SubAgentTool

__all__ = [
    "ReadFileTool",
    "WriteFileTool",
    "EditFileTool",
    "ShellTool",
    "WebSearchTool",
    "MemoryTool",
    "TodosTool",
    "BrowserUseTool",
    "VisionCaptureTool",
    "MCPExecutorTool",
    "SubAgentTool",
]

def get_all_builtin_tools() -> List[Tool]:
    return [
        ReadFileTool,
        WriteFileTool,
        ShellTool,
        EditFileTool,
        WebSearchTool,
        MemoryTool,
        TodosTool,
        BrowserUseTool,
        VisionCaptureTool,
        MCPExecutorTool,
        SubAgentTool,
    ]



from dataclasses import dataclass, field
from typing import Dict, Any, Union
from enum import Enum
from pegasus.utils.logger import logger
import json
from typing import List

class StreamEventType(str, Enum):
    TEXT_DELTA = 'text_delta'
    TOOL_CALL_COMPLETE = 'tool_call_complete'
    MESSAGE_COMPLETE = 'message_complete' # always has usage data
    ERROR = 'error'

@dataclass
class TextDelta:
    content: str

    def __add__(self, other: 'TextDelta') -> 'TextDelta':
        return TextDelta(content=self.content + other.content)

    def __str__(self) -> str:
        return self.content


@dataclass
class ReasoningDelta:
    reasoning: str

    def __add__(self, other: 'ReasoningDelta') -> 'ReasoningDelta':
        return ReasoningDelta(reasoning=self.reasoning + other.reasoning)

    def __str__(self) -> str:
        return self.reasoning
    
@dataclass
class TokenUsage:
    prompt_tokens: int = 0
    completion_tokens: int = 0
    total_tokens: int = 0
    total_time: float | None = None  # seconds

    def __add__(self, other: 'TokenUsage') -> 'TokenUsage':
        return TokenUsage(
            prompt_tokens=self.prompt_tokens + other.prompt_tokens,
            completion_tokens=self.completion_tokens + other.completion_tokens,
            total_tokens=self.total_tokens + other.total_tokens,
        )

    def __sub__(self, other: 'TokenUsage') -> 'TokenUsage':
        return TokenUsage(
            prompt_tokens=self.prompt_tokens - other.prompt_tokens,
            completion_tokens=self.completion_tokens - other.completion_tokens,
            total_tokens=self.total_tokens - other.total_tokens,
        )

@dataclass
class ToolCallDelta:
    call_id: str
    name: str | None = None
    arguments_delta: str = ""

@dataclass
class ToolCall:
    call_id: str
    name: str | None = None
    arguments: Dict[str, Any] = field(default_factory=dict)

@dataclass
class StreamEvent:
    type: StreamEventType
    text_delta: TextDelta | None = None
    reasoning_delta: ReasoningDelta | None = None
    refusal: str | None = None
    annotations: Dict[str, Any] = field(default_factory=dict)
    error: str | None = None
    finish_reason: str | None = None
    usage: TokenUsage | None = None
    tool_calls: List[ToolCall] = field(default_factory=list)

def merge_text_deltas(deltas: List[TextDelta]) -> TextDelta | None:
    """given a list of text deltas, add all content together."""
    parts = [d.content for d in deltas if d.content]
    return TextDelta(content="".join(parts)) if parts else None


def merge_reasoning_deltas(deltas: List[ReasoningDelta]) -> ReasoningDelta | None:
    """given a list of reasoning deltas, add all reasoning together."""
    parts = [d.reasoning for d in deltas if d.reasoning]
    return ReasoningDelta(reasoning="".join(parts)) if parts else None


def parse_tool_call_arguments(arguments: Union[str, Dict, None]) -> Dict[str, Any]:
    try:
        if not arguments:
            return {}
        if isinstance(arguments, str):
            return json.loads(arguments)
        return arguments
    except Exception as e:
        logger.error(f"Error parsing tool call arguments: {e} with arguments: {arguments}")
        return {}


from pegasus.client.llm_client import LLMClient
from pegasus.client.response import (
    StreamEvent,
    StreamEventType,
    TextDelta,
    ReasoningDelta,
    TokenUsage,
    ToolCall,
    ToolCallDelta,
    merge_text_deltas,
    merge_reasoning_deltas,
    parse_tool_call_arguments,
)

__all__ = [
    "LLMClient",
    "StreamEvent",
    "StreamEventType",
    "TextDelta",
    "ReasoningDelta",
    "TokenUsage",
    "ToolCall",
    "ToolCallDelta",
    "merge_text_deltas",
    "merge_reasoning_deltas",
    "parse_tool_call_arguments",
]



from pegasus.config.config import Config
from pegasus.context.compaction import ChatCompressor
from pegasus.context.manager import ContextManager
from pegasus.client.llm_client import LLMClient
from pegasus.mcp.manager import MCPManager
from pegasus.tools.registry import create_default_registry, ToolRegistry
import uuid
from datetime import datetime

class Session:
    def __init__(self, config: Config, tool_registry: ToolRegistry | None = None) -> None:
        self.session_id = str(uuid.uuid4())
        self.config = config
        self.client = LLMClient(self.config)
        self.chat_compressor = ChatCompressor(self.client)
        self.context_manager: ContextManager | None = None
        self.tool_registry: ToolRegistry | None = tool_registry
        self.mcp_manager = MCPManager(self.config)
        self.config.mcp_manager = self.mcp_manager
        self.created_at = datetime.now()
        self.updated_at = datetime.now()
        self.turn_count = 0

    async def initialize(self) -> None:
        await self.mcp_manager.initialize()
        if self.tool_registry is None:
            self.tool_registry = create_default_registry(self.config)
        self.context_manager = ContextManager(self.config)

    async def increment_turn(self) -> None:
        self.turn_count += 1
        self.updated_at = datetime.now()
        return self.turn_count
    

    async def get_stats(self) -> dict[str, str]:
        return {
            "turn_count": self.turn_count,
            "updated_at": self.updated_at.isoformat(),
            "created_at": self.created_at.isoformat(),

        }

    async def cleanup(self) -> None:
        if self.tool_registry is not None:
            await self.tool_registry.cleanup()
        await self.mcp_manager.shutdown()
        self.config.mcp_manager = None
        await self.client.close()
    
    


"""
Pegasus - A lightweight terminal agent framework
"""

from pegasus.agent.agent import Agent
from pegasus.agent.results import AgentEvent, AgentEventType
from pegasus.config.config import Config, ModelConfig
from pegasus.session.session import Session
from pegasus.cli.cli import run_cli

__version__ = "0.1.6"
__all__ = [
    "Agent",
    "AgentEvent",
    "AgentEventType",
    "Config",
    "ModelConfig",
    "Session",
    "run_cli",
]



from pegasus.tools.base import Tool
from pegasus.tools.data import ToolType, ToolInvocation, ToolResult
from pegasus.config.config import Config
from pydantic import BaseModel, Field
from enum import Enum
from pathlib import Path
import json

class MemoryAction(str, Enum):
    SET = "set"
    GET = "get"
    LIST = "list"
    DELETE = "delete"
    CLEAR = "clear"
    def __str__(self) -> str:
        return self.value

class MemoryParams(BaseModel):
    action: str = Field(..., description="The action to perform on the memory. Actions are: set, get, list, delete, clear")
    key: str = Field(..., description="The key of the memory to store and retrieve. Used in the set and get actions")
    value: str = Field(..., description="The value of the memory to store and retrieve. Used in the set action")

class MemoryTool(Tool):
    name: str = "memory"
    description: str = "Manage memory for the current session. Use this to store and retrieve important information that you want to remember. "\
                       "The memory tool can have information from previous sessions."\
                       " If asked about user specific information or need specific user information to perform tasks check if it is available in the memory."\
                       "When not certain list out all the keys in the memory to help you decide which key to use."
    type: ToolType = ToolType.MEMORY
    schema: MemoryParams = MemoryParams

    def __init__(self, config: Config) -> None:
        self._config = config
        self.memory_file = Path(self._config.cwd / "pegasus_memory.json") # TODO: this should be stored in the platformdir config directory to make it persistent across sessions
        self._memory: dict[str, str] = {}
        if not self.memory_file.exists():
            self.memory_file.touch()
            with open(self.memory_file, "w") as f:
                json.dump({}, f)
        else:
            with open(self.memory_file, "r") as f:
                self._memory = json.load(f)

    def _persist(self) -> None:
        with open(self.memory_file, "w") as f:
            json.dump(self._memory, f, indent=4)

    def _execute_set(self, params: MemoryParams) -> ToolResult:
        if not params.key or not params.value:
            return ToolResult.error_result("memory key and value are required for the set action")
        self._memory[params.key] = params.value
        self._persist()
        return ToolResult.success_result(output=f"memory stored: {params.key} = {params.value}")

    def _execute_get(self, params: MemoryParams) -> ToolResult:
        if not params.key:
            return ToolResult.error_result("key is required for the get action")
        if params.key not in self._memory:
            return ToolResult.error_result(f"memory with key: {params.key} not found")
        return ToolResult.success_result(output=f"memory retrieved: {self._memory[params.key]}")

    def _execute_list(self, params: MemoryParams) -> ToolResult:
        output_lines = [f"[{k}]: {v}" for k, v in self._memory.items()]
        if not output_lines:
            return ToolResult.success_result(output="no memory found. use the set action to create memory.")
        return ToolResult.success_result(output="\n".join(output_lines))

    def _execute_delete(self, params: MemoryParams) -> ToolResult:
        if not params.key:
            return ToolResult.error_result("key is required for the delete action")
        if params.key not in self._memory:
            return ToolResult.error_result(f"memory with key: {params.key} not found")
        del self._memory[params.key]
        self._persist()
        return ToolResult.success_result(output=f"memory deleted: {params.key}")

    def _execute_clear(self, params: MemoryParams) -> ToolResult:
        self._memory = {}
        self._persist()
        return ToolResult.success_result(output="cleared all memory")

    async def _execute(self, invocation: ToolInvocation) -> ToolResult:
        params = MemoryParams(**invocation.params)
        if params.action == MemoryAction.SET:
            return self._execute_set(params)
        if params.action == MemoryAction.GET:
            return self._execute_get(params)
        if params.action == MemoryAction.LIST:
            return self._execute_list(params)
        if params.action == MemoryAction.DELETE:
            return self._execute_delete(params)
        if params.action == MemoryAction.CLEAR:
            return self._execute_clear(params)
        return ToolResult.error_result(
            f"invalid action: {params.action}. valid actions are: {', '.join(a.value for a in MemoryAction)}"
        )


    async def execute(self, invocation: ToolInvocation) -> ToolResult:
        try:
            return await self._execute(invocation)
        except Exception as e:
            return ToolResult.error_result(f"Error managing memory with error: {e}")


from pegasus.tools.base import Tool
from pegasus.tools.data import ToolType, ToolInvocation, ToolResult
from pegasus.config.config import Config
from pydantic import BaseModel, Field
from pathlib import Path
import asyncio
import fnmatch
import sys
import os

class ShellParams(BaseModel):
    command: str = Field(..., description="The command to run")
    cwd: str = Field(None, description="The current working directory to run the command in. If not provided, the current working directory will be used.")
    timeout: int = Field(120, ge=1, le=600, description="The timeout in seconds for the command (default 120 seconds, max 600 seconds, min 1 second)")


BLOCKED_COMMANDS = {
    "rm -rf /",
    "rm -rf ~",
    "rm -rf /*",
    "dd if=/dev/zero",
    "dd if=/dev/random",
    "mkfs",
    "fdisk",
    "parted",
    ":(){ :|:& };:",
    "chmod 777 /",
    "chmod -R 777",
    "shutdown",
    "reboot",
    "halt",
    "poweroff",
    "init 0",
    "init 6",
}

class ShellTool(Tool):
    name: str = "shell"
    description: str = "Execute a shell command. Use this for running commands, scripts and cli tools"
    type: ToolType = ToolType.BASH
    schema: ShellParams = ShellParams

    def __init__(self, config: Config) -> None:
        self._config = config

    def _build_environment(self) -> dict[str, str]:
        env = os.environ.copy()
        shell_environment = self._config.shell_environment
        # remove default excludes
        if shell_environment.ignore_default_excludes:
            for pattern in shell_environment.exclude_patterns:
                keys_to_remove = [k for k in env.keys() if fnmatch.fnmatch(k.upper(), pattern.upper())]
                for key in keys_to_remove:
                    del env[key]
        # add set vars
        if shell_environment.set_vars:
            env.update(shell_environment.set_vars)
        return env

    async def _execute(self, invocation: ToolInvocation) -> ToolResult:
        params = ShellParams(**invocation.params)
        if params.command.lower().strip() in BLOCKED_COMMANDS:
            return ToolResult.error_result(f"Command is blocked: {params.command}", metadata=dict(blocked=True))
        if params.cwd:
            cwd = Path(params.cwd)
            if not cwd.is_absolute():
                cwd = invocation.cwd / cwd
        else:
            cwd = invocation.cwd
        
        if not cwd.exists():
            return ToolResult.error_result(f"Current working directory does not exist: {cwd}", metadata=dict(cwd=str(cwd.as_posix())))

        env = self._build_environment()
        if sys.platform == "win32":
            command = ["cmd.exe", "/c", params.command]
        else:
            command = ["bash", "-c", params.command]

        process = await asyncio.create_subprocess_exec(*command, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE, cwd=cwd, env=env, start_new_session=True)
        try:
            stdout, stderr = await asyncio.wait_for(process.communicate(), timeout=params.timeout)
        except asyncio.TimeoutError:
            if sys.platform == "win32":
                os.killpg(os.getpgid(process.pid), signal.SIGTERM)
            else:
                process.kill()
            return ToolResult.error_result(f"command timed out: {params.command} in {params.timeout} seconds", metadata=dict(timeout=params.timeout))

        stdout = stdout.decode("utf-8", errors="replace").strip()
        stderr = stderr.decode("utf-8", errors="replace").strip()
        exit_code = process.returncode
        
        output = ""
        truncated = False
        if stdout.strip():
            output += stdout.rstrip()
        if stderr.strip():
            output += "\n--- stderr ---\n" + stderr.rstrip()
        if exit_code != 0:
            output += f"\n--- exit code ---\n{exit_code}"

        if len(output) > 100 * 1024:
            output = output[:100 * 1024] + "\n...truncated..."
            truncated = True
        
        return ToolResult(
            success=exit_code == 0,
            error=stderr if exit_code != 0 else None,
            exit_code=exit_code,
            output=output,
            truncated=truncated,
        )


    async def execute(self, invocation: ToolInvocation) -> ToolResult:
        try:
            return await self._execute(invocation)
        except Exception as e:
            return ToolResult.error_result(f"Error executing shell tool with error: {e}")


from __future__ import annotations

from typing import Dict, List, Any, Type
from pegasus.tools.base import Tool
from pegasus.utils.logger import logger
from pegasus.tools.data import ToolResult, ToolInvocation
from pegasus.tools.builtin import get_all_builtin_tools
from pegasus.config.config import Config
from pathlib import Path


class ToolRegistry:
    def __init__(self, config: Config):
        self._tools: Dict[str, Tool] = {}
        self._config = config

    def register(self, tool: Tool) -> None:
        if tool.name in self._tools:
            logger.warning(f"[tool registry] tool {tool.name} already registered. overwriting...")
        self._tools[tool.name] = tool
        logger.info(f"[tool registry] tool {tool.name} activated🔥")


    def unregister(self, tool_name: str) -> bool:
        if tool_name not in self._tools:
            logger.warning(f"tool {tool_name} not found in registry")
            return False
        del self._tools[tool_name]
        logger.debug(f"tool {tool_name} unregistered successfully")
        return True
    
    def get_schemas(self) -> List[dict[str, Any]]:
        tools = self.get_all()
        return [tool.to_openai_schema() for tool in tools]

    def get(self, tool_name: str) -> Tool | None:
        if tool_name in self._tools:
            return self._tools[tool_name]
        return None
    
    def get_all(self) -> List[Tool]:
        return list(self._tools.values())
    
    async def invoke(self, name: str, params: Dict[str, Any], cwd: Path) -> ToolResult:
        tool: Tool | None = self.get(name)
        if tool is None:
            return ToolResult.error_result(f"tool {name} not found in registry")

        validation_errors = tool.validate_params(params)
        if validation_errors:
            return ToolResult.error_result(f"validation errors: {'; '.join(validation_errors)}", metadata=dict(tool_name=name, validation_errors=validation_errors))
        tool_invocation = ToolInvocation(cwd=cwd, params=params)
        try:
            result = await tool.execute(tool_invocation)
            return result
        except Exception as e:
            logger.exception(f"error invoking tool {name} with error: {e}")
            return ToolResult.error_result(f"internal error while invoking tool {name}: {e}")

    async def cleanup(self) -> None:
        seen_ids: set[int] = set()
        for tool in self.get_all():
            tool_id = id(tool)
            if tool_id in seen_ids:
                continue
            seen_ids.add(tool_id)
            close = getattr(tool, "close", None)
            if close is None:
                continue
            try:
                await close()
            except Exception as e:
                logger.warning(f"error cleaning up tool {tool.name}: {e}")
        

def create_default_registry(config: Config) -> ToolRegistry:
    registry = ToolRegistry(config)
    BUILTIN_TOOLS: List[Type[Tool]] = get_all_builtin_tools()
    for tool in BUILTIN_TOOLS:
        registry.register(tool(config))
    return registry
    



import tomli
from pathlib import Path
from typing import Any
from pegasus.config.config import Config
from platformdirs import user_config_dir
from pegasus.utils.errors import ConfigError
from pegasus.utils.logger import logger

_CONFIG_FILE_NAME = "config.json"

def get_config_dir() -> Path:
    return Path(user_config_dir("pegasus", appauthor=False))

def get_system_config_path() -> Path:
    return get_config_dir() / _CONFIG_FILE_NAME

def _parse_toml(path: Path):
    try:
        with open(path, "rb") as f:
            return tomli.load(f)
    except tomli.TOMLDecodeError as e:
        raise ConfigError(
            message=f"Invalid TOML file in {path}: {e}",
            config_file=path,
            cause=e
        )
    except (OSError, IOError) as e:
        raise ConfigError(
            message=f"Failed to read TOML file in {path}: {e}",
            config_file=path,
            cause=e
        )

def _get_project_config(cwd: Path) -> Path:
    current = cwd.resolve()
    agent_dir = current / ".pegasus"
    if agent_dir.is_dir():
        config_path = agent_dir / "config.toml"
        if config_path.is_file():
            return config_path
    return None


def _merge_dicts(base: dict[str, Any], override: dict[str, Any]) -> dict[str, Any]:
    result = base.copy()
    for key, value in override.items():
        if key in result and isinstance(result[key], dict) and isinstance(value, dict):
            result[key] = _merge_dicts(result[key], value)
        else:
            result[key] = value
    return result

def load_config(cwd: Path | None) -> Config:
    cwd = cwd or Path.cwd()
    system_path = get_system_config_path()
    config_dict: dict[str, Any] = {}
    if system_path.is_file():
        try:
           config_dict = _parse_toml(system_path)
        except ConfigError as e:
            logger.warning(f"Failed to load system config: {e}")
    project_path = _get_project_config(cwd)
    if project_path:
        try:
            project_dict = _parse_toml(project_path)
            config_dict = _merge_dicts(config_dict, project_dict)
        except ConfigError as e:
            logger.warning(f"Failed to load project config: {e}")

    if "cwd" not in config_dict:
        config_dict["cwd"] = cwd

    try:
        config = Config(**config_dict)
    except Exception as e:
        raise ConfigError(
            message=f"Invalid config: {e}",
            config_file=system_path,
            cause=e
        ) from e
    return config


from rich.console import Console
from rich.theme import Theme
from rich.rule import Rule
from rich.text import Text
from typing import Dict, Any, List, Tuple, Set
from rich.table import Table
from rich import box
from rich.panel import Panel
from pathlib import Path
from rich.syntax import Syntax
from rich.console import Group
import re
from pegasus.utils.paths import display_path_rel_to_cwd
from pegasus.config.config import Config
from pegasus.utils.text import truncate_text
from pegasus.tools.data import FileDiff
from pegasus.utils.logger import logger
from art import text2art

_GREEN = "\033[32m"
_RESET = "\033[0m"

# flying horse (Pegasus) ascii art
FLYING_HORSE_ART = r"""
====================================================================================================
====================================================================================================
=====-:=====================================================================================-:-=====
=====-.:-===============================================-==================================-..-=====
======:..:-===============================:.===========- :===============================-:..:======
===--==:...:=============================.:=:=========-:*.:============================-:...:==--===
===:.:-=-:...:-========================== +#::========.+#::==========================-:...:---:.-===
====:...:-::...::-=======================::+:..........:-.-=======================--:...::-:...:====
=====:.....:::......-=====================.     ....     .=====================-.....::::....::=====
==-.::--::....:::.    .:--===============:    ........    :===============--:.    .:::....::--::.-==
===-....:::::....         .-=+++=========:   ..........   -=========+++=-.         ....:::::....-===
=====:::......:::             .-*#++======   ..........   ======+*#*-.             :::......:::=====
====------::::....               :+##+===-%* .......... +%-===+##+.               ....::::------====
=====:......::::::                 .=###+-=*: ........ .+=-+###=.                 ::::::......:=====
======--:::........                  .-*#++-   .......  :++#+-.                 .........:::--======
=======::::::::......                 .=+++-   ......   .+++=.                 ......::::::::=======
========-::.::::::::..              .-+++++:   ......   .+++++:               ..::::::::.::-========
============::.....:::...         .:+++++++.   ......    =+++++=.          ...:::.....::============
=============-:::::...::....     .=++++++++..  ......   .=+++++++-.     ....::...:::::-=============
===============-:..:::..::..::..=++++++++++..  ......  ..-++++++++=:..::..::..:::..:-===============
===================:..::...:::-+++++++++++=... ......  ..-++++++++++=:.::...::..:===================
====================-=...::::+++++++++++++-... ...... ...:++++++++++++:.:::..:=-====================
========================--:=++++++++++++++-....:....:=...:+++++++++++++=::--========================
=========================+++++++++++++++++:...=@+...#@:...+++++++++++++++===========================
=======================+++++++++++++++++++:...:%%..:@+....=++++++++++++++++=========================
======================++++++++++++++++++++....:-:..:=*:...=++++++++++++++++++=======================
====================+++++++++++++++++++++=.....:----::....-+++++++++++++++++++======================
==================+++++++++++++++++++++++=................:+++++++++++++++++++++====================
"""

AGENT_THEME = Theme(
    {
        "info": "cyan",
        "warning": "yellow",
        "error": "bright_red bold",
        "success": "green",
        "dim": "dim",
        "muted": "gray50",
        "border": "grey35", 
        "highlight": "bold cyan",
        "reasoning": "bright_green",
        # roles
        "user": "bright_blue bold",
        "assistant": "blue bold",
        # tools
        "tool": "bright_magenta bold",
        "tool.read": "cyan",
        "tool.write": "yellow",
        "tool.edit": "bright_yellow",
        "tool.bash": "magenta",
        "tool.network": "bright_blue",
        "tool.memory": "green",
        "tool.mcp": "bright_cyan",
        "tool.sub_agent": "bright_magenta",
        # code / blocks
        "code": "white",
    }
)

_console: Console | None = None

def get_console() -> Console:
    global _console
    if _console is None:
        _console = Console(theme=AGENT_THEME, highlight=True)
    return _console

class TUI:
    def __init__(self, config: Config, console: Console | None = None) -> None:
        self._console = console or get_console()
        self._config = config
        self._assistant_stream_open = False
        self._reasoning_stream_open = False
        self._tool_arguments_by_call_id: Dict[str, Dict[str, Any]] = {}
        self._cwd: Path | None = config.cwd
        self._full_width: int | None = self._console.width

    def _begin_assistant(self) -> None:
        self._console.print()
        self._console.print(Rule(Text("Assistant End", style="assistant")))
        self._assistant_stream_open = True

    def _end_assistant(self) -> None:
        self._console.print()
        self._console.print(Rule(Text("Assistant Start", style="assistant")))
        self._assistant_stream_open = False
    
    def stream_assistant_delta(self, content: str) -> None:
        self._console.print(content, end="", markup=False, style="assistant") 
    
    def _begin_reasoning(self) -> None:
        self._console.print()
        self._console.print(Rule(Text("Reasoning Start", style="reasoning")))
        self._reasoning_stream_open = True

    def stream_reasoning_delta(self, content: str) -> None:
        self._console.print(content, end="", markup=False, style="reasoning")
    
    def _end_reasoning(self) -> None:
        self._console.print()
        self._console.print(Rule(Text("Reasoning End", style="reasoning")))
        self._reasoning_stream_open = False
    
    def _order_args(self, tool_name:str, args: dict[str, Any]) -> list[str]:
        _PREFERRED_ORDER= {
            "read_file": ["path", "offset", "limit"],
            "write_file": ["path", "create_dirs", "content"],
            "edit_file": ["path", "replace_all", "old_string", "new_string"],
            "shell": ["command", "timeout", "cwd"],
            "list_dir": ["path", "include_hidden"],
            "grep": ["path", "pattern", "case_sensitive"],
            "glob": ["path", "pattern"],
            "memory": ["action", "key", "value"],
            "todos": ["action", "id", "content"],
        }
        preferred_order = _PREFERRED_ORDER.get(tool_name, [])
        ordered: List[Tuple[str, Any]] = []
        seen: Set[str] = set()
        for key in preferred_order:
            if key in args:
                ordered.append((key, args[key]))
                seen.add(key)
        remaining_keys = set(args.keys() - seen)
        ordered.extend((key, args[key]) for key in remaining_keys)
        return ordered



    def _render_arguments_table(self, tool_name: str, arguments: dict[str, Any]) -> str:
        table = Table.grid(padding=(0, 1))
        table.add_column(style='muted', justify='right', no_wrap=True)
        table.add_column(style='code', overflow='fold')
        for key, value in self._order_args(tool_name, arguments):
            if isinstance(value, str) and key in { "content", "old_string", "new_string", "old_content", "new_content"}:
                line_count = len(value.splitlines()) or 0
                byte_count = len(value.encode("utf-8")) or 0
                value = f"<{line_count} lines - {byte_count} bytes>"
            if isinstance(value, bool):
                value = "true" if value else "false"
            table.add_row(key, str(value))
        return table

    def _guess_lexer(self, path: str | None) -> str:
        if not path:
            return "text"
        suffix = Path(path).suffix.lower()
        if suffix in (".py", ".pyw", ".pyx", ".pyz"):
            return "python"
        if suffix in (".js", ".jsx", ".ts", ".tsx"):
            return "javascript"
        if suffix in (".html", ".htm", ".xhtml", ".xht"):
            return "html"
        if suffix == ".css":
            return "css"
        return "text"

    def tool_call_start(
            self,
            call_id: str,
            name: str, 
            tool_kind: str | None, 
            arguments: dict[str, Any]
    ) -> None:
        self._tool_arguments_by_call_id[call_id] = arguments
        border_style = f"tool.{tool_kind.value}" if tool_kind else "tool"

        title = Text.assemble(("🤔 ", "muted"), (name, "tool"), (" ", "muted"), (f"#{call_id[:8]}", "muted"))
        display_args = dict(arguments)
        for key in ('pwd', 'cwd'):
            val = display_args.get(key)
            if isinstance(val, str) and self._cwd:
                display_args[key] = str(display_path_rel_to_cwd(val, self._cwd))
        panel = Panel(self._render_arguments_table(name, arguments) if display_args else Text("No arguments", style="muted"), 
                      title=title, 
                      title_align="left",
                      subtitle=Text('running...', style="muted"),
                      subtitle_align="right",
                      box=box.ROUNDED,
                      padding=(1, 2),
                      border_style=border_style)
        self._console.print()
        self._console.print(panel)

    def _extract_read_file_code(self, text: str) -> tuple[int, str] | None:
        body = text
        header_match = re.match(r"^Showing lines (\d+)-(\d+) of (\d+)\n\n", body)
        if header_match:
            body = text[header_match.end():]

        code_lines: list[str] = []
        start_line: int | None = None
        for line in body.splitlines():
            m = re.match(r"^\s*(\d+)\|(.*)$", line)
            if not m:
                return None
            line_number = int(m.group(1))
            if start_line is None:
                start_line = line_number
            code_line = m.group(2)
            code_lines.append(code_line)
        if start_line is None:
            return None
        return start_line, "\n".join(code_lines)

    def _render_read_file_complete(
        self,
        output: str,
        primary_path: str | None,
        shown_start: int | None,
        shown_end: int | None,
        total_lines: int | None,
    ) -> List[Any]:
        blocks = []
        parsed = self._extract_read_file_code(output)
        if not parsed:
            blocks.append(Syntax(output, "text", theme="monokai", word_wrap=False))
            return blocks
        start_line, code = parsed
        lexer = self._guess_lexer(primary_path)
        header_parts = [display_path_rel_to_cwd(primary_path, self._cwd) if primary_path else "unknown file"]
        header_parts.append(" . ")
        if shown_start is not None and shown_end is not None and total_lines is not None:
            header_parts.append(f"Showing lines {shown_start}-{shown_end} of {total_lines}")
        blocks.append(Text(" ".join(header_parts), style="muted"))
        blocks.append(Syntax(code, lexer, theme="monokai", line_numbers=True, start_line=start_line, word_wrap=False))
        return blocks

    def _render_write_edit_complete(self, output: str, diff: FileDiff | None, truncated: bool) -> List[Any]:
        blocks = []
        output_line = output.strip() if output.strip() else "Completed"
        blocks.append(Text(output_line, style="success"))
        diff_text = diff.create_diff() if diff else ""
        diff_display = truncate_text(diff_text, self._config.model_name, self._config.max_tool_output_tokens)
        blocks.append(Syntax(diff_display, "diff", theme="monokai", line_numbers=True, word_wrap=False))
        if truncated:
            blocks.append(Text("Output truncated", style="warning"))
        return blocks

    def _render_shell_complete(
        self, args: dict[str, Any], output: str, exit_code: int | None, truncated: bool
    ) -> List[Any]:
        blocks = []
        command = args.get("command")
        if isinstance(command, str) and command.strip():
            blocks.append(Text(f"${command}", style="muted"))
        if exit_code is not None:
            blocks.append(Text(f"exit code: {exit_code}", style="muted"))
        output_display = truncate_text(output, self._config.model_name, self._config.max_tool_output_tokens)
        blocks.append(Syntax(output_display, "text", theme="monokai", word_wrap=False))
        if truncated:
            blocks.append(Text("Output truncated", style="warning"))
        return blocks

    def _render_list_dir_complete(self, output: str, metadata: dict[str, Any], truncated: bool) -> List[Any]:
        blocks = []
        path = metadata.get("path")
        entries = metadata.get("entries")
        summary = []
        if isinstance(path, str):
            summary.append(path)
        if isinstance(entries, int):
            summary.append(f"{entries} entries")
        if summary:
            blocks.append(Text(" ".join(summary), style="muted"))
        output_display = truncate_text(output, self._config.model_name, self._config.max_tool_output_tokens)
        blocks.append(Syntax(output_display, "text", theme="monokai", word_wrap=False))
        if truncated:
            blocks.append(Text("Output truncated", style="warning"))
        return blocks

    def _render_grep_complete(self, output: str, truncated: bool) -> List[Any]:
        blocks = []
        output_display = truncate_text(output, self._config.model_name, self._config.max_tool_output_tokens)
        blocks.append(Syntax(output_display, "text", theme="monokai", word_wrap=False))
        if truncated:
            blocks.append(Text("Output truncated", style="warning"))
        return blocks

    def _render_glob_complete(self, output: str, truncated: bool) -> List[Any]:
        blocks = []
        output_display = truncate_text(output, self._config.model_name, self._config.max_tool_output_tokens)
        blocks.append(Syntax(output_display, "text", theme="monokai", word_wrap=False))
        if truncated:
            blocks.append(Text("Output truncated", style="warning"))
        return blocks

    def _render_websearch_complete(self, output: str, metadata: dict[str, Any], truncated: bool) -> List[Any]:
        blocks = []
        results = metadata.get("results")
        query = metadata.get("query")
        summary = []
        if isinstance(query, list):
            summary.append(f"Search results for '{query}'")
        if isinstance(results, list):
            summary.append(f"Total results: {len(results)}")
        if isinstance(results, int):
            summary.append(f"{results} results")
        if summary:
            blocks.append(Text(".".join(summary), style="muted"))
        output_display = truncate_text(output, self._config.model_name, self._config.max_tool_output_tokens)
        blocks.append(Syntax(output_display, "text", theme="monokai", word_wrap=False))
        if truncated:
            blocks.append(Text("Output truncated", style="warning"))
        return blocks

    def _render_todos_complete(self, output: str, truncated: bool) -> List[Any]:
        blocks = []
        output_display = truncate_text(output, self._config.model_name, self._config.max_tool_output_tokens)
        blocks.append(Syntax(output_display, "text", theme="monokai", word_wrap=False))
        if truncated:
            blocks.append(Text("Output truncated", style="warning"))
        return blocks

    def _render_memory_complete(self, output: str, truncated: bool) -> List[Any]:
        blocks = []
        output_display = truncate_text(output, self._config.model_name, self._config.max_tool_output_tokens)
        blocks.append(Syntax(output_display, "text", theme="monokai", word_wrap=False))
        if truncated:
            blocks.append(Text("Output truncated", style="warning"))
        return blocks

    def _render_error_blocks(self, error: str | None, output: str) -> List[Any]:
        blocks = []
        if error:
            blocks.append(Text(error, style="error"))
        output_display = truncate_text(output, self._config.model_name, self._config.max_tool_output_tokens)
        if output_display.strip():
            blocks.append(Syntax(output_display, "text", theme="monokai", word_wrap=False))
        else:
            blocks.append(Text("No output", style="error"))
        return blocks

    def _render_tool_complete_blocks(
        self,
        name: str,
        call_id: str,
        success: bool,
        output: str,
        error: str | None,
        metadata: dict[str, Any],
        truncated: bool,
        diff: FileDiff | None,
        exit_code: int | None,
    ) -> List[Any]:
        args = self._tool_arguments_by_call_id.get(call_id, {})
        meta = metadata if isinstance(metadata, dict) else {}
        primary_path = meta.get("path") if isinstance(meta.get("path"), str) else None
        shown_start = meta.get("shown_start")
        shown_end = meta.get("shown_end")
        total_lines = meta.get("total_lines")

        if not success:
            return self._render_error_blocks(error, output)
        if name == "read_file":
            return self._render_read_file_complete(output, primary_path, shown_start, shown_end, total_lines)
        if name in ("write_file", "edit_file"):
            return self._render_write_edit_complete(output, diff, truncated)
        if name == "shell":
            return self._render_shell_complete(args, output, exit_code, truncated)
        if name == "list_dir":
            return self._render_list_dir_complete(output, meta, truncated)
        if name == "grep":
            return self._render_grep_complete(output, truncated)
        if name == "glob":
            return self._render_glob_complete(output, truncated)
        if name == "websearch":
            return self._render_websearch_complete(output, meta, truncated)
        if name == "todos":
            return self._render_todos_complete(output, truncated)
        if name == "memory":
            return self._render_memory_complete(output, truncated)

        output_display = truncate_text(output, self._config.model_name, self._config.max_tool_output_tokens)
        blocks = [Syntax(output_display, "text", theme="monokai", word_wrap=False)] if output_display.strip() else []
        image_paths = meta.get("image_paths")
        if isinstance(image_paths, list) and image_paths:
            blocks.insert(0, Text(f"images: {len(image_paths)}", style="muted"))
        return blocks

    def tool_call_complete(
                self,
                call_id: str,
                name: str, 
                tool_kind: str | None, 
                success: bool,
                output: str,
                error: str | None,
                metadata: dict[str, Any],
                truncated: bool,
                diff: FileDiff | None,
                exit_code: int | None,
        ) -> None:
            border_style = f"tool.{tool_kind.value}" if tool_kind else "tool"
            status_icon = "😊" if success else "😭"
            status_style = "success" if success else "error"
            title = Text.assemble((f"{status_icon} ", status_style), (name, "tool"), (" ", "muted"), (f"#{call_id[:8]}", "muted"))
            blocks = self._render_tool_complete_blocks(name, call_id, success, output, error, metadata, truncated, diff, exit_code)
            panel = Panel(
                Group(*blocks),
                title=title,
                title_align="left",
                subtitle=Text("completed" if success else "failed", style="muted"),
                subtitle_align="right",
                box=box.ROUNDED,
                padding=(1, 2),
                border_style=border_style,
            )
            self._console.print()
            self._console.print(panel)

    def print_help(self) -> None:
        self._console.print(Panel(Text("Help", style="highlight"), title="Help", title_align="left", box=box.ROUNDED, padding=(1, 2), border_style="border"))
        self._console.print(Text("Commands:", style="muted"))
        self._console.print(Text("  /help - Show this help message", style="muted"))
        self._console.print(Text("  /exit - Exit the program", style="muted"))
        self._console.print(Text("  /quit - Exit the program", style="muted"))
        self._console.print(Text("  /config - Show the current configuration", style="muted"))
        self._console.print(Text("  /model - Show the current model", style="muted"))


    def print_welcome(self, title: str, lines: List[str]) -> None:
        # flying horse at top (green), no markup so brackets in art are preserved
        horse_text = Text.from_ansi(f"{_GREEN}{FLYING_HORSE_ART.strip()}{_RESET}")
        self._console.print(horse_text)
        self._console.print()
        # "Pegasus" in smaller font so it fits the console
        try:
            pegasus_art = text2art("Pegasus - Gives  you  wings!", font="medium")
        except Exception:
            pegasus_art = text2art("Pegasus - Gives  you  wings!")
        pegasus_text = Text.from_ansi(f"{_GREEN}{pegasus_art}{_RESET}")
        self._console.print(pegasus_text)
        self._console.print()
        body = "\n".join(lines)
        self._console.print(
            Panel(
                Text(body, style="code"),
                title=Text(title, style="highlight"),
                title_align="left",
                border_style="border",
                box=box.ROUNDED,
                padding=(1, 2),
                width=self._full_width,
                style="on grey23",
            )
        )

    def print_config(self) -> None:
        self._console.print(Panel(Text("Configuration", style="highlight"), title="Configuration", title_align="left", box=box.ROUNDED, padding=(1, 2), border_style="border"))
        self._console.print(Text(f"model: {self._config.model_name}", style="muted"))
        self._console.print(Text(f"cwd: {self._config.cwd}", style="muted"))
        # self._console.print(Text(f"max_tool_output_tokens: {self._config.max_tool_output_tokens}", style="muted"))
        # self._console.print(Text(f"max_tool_output_tokens: {self._config.max_tool_output_tokens}", style="muted"))


from typing import List, Dict, Any
from pegasus.context.data import MessageItem, MessageRole, ToolResultMessage
from pegasus.prompts.system import get_system_prompt
from pegasus.utils.text import count_tokens
from pegasus.config.config import Config
from pegasus.client.response import TokenUsage
from pegasus.tools.data import ToolImage


class ContextManager:
    PRUNE_PROTECT_TOKENS = 40_000
    PRUNE_MINIMUM_TOKENS = 20_000
    
    def __init__(self, config: Config) -> None:
        self.config = config
        self.system_prompt = get_system_prompt(config)
        self._messages: List[MessageItem] = []
        self.model_name: str | None = config.model_name
        self.current_usage: TokenUsage = TokenUsage()
        self.latest_usage: TokenUsage = TokenUsage()
        self.total_usage: TokenUsage = TokenUsage()

    def _message_token_count(self, content: str) -> int:
        return count_tokens(content or "", self.model_name or "")

    def _refresh_current_usage(self) -> None:
        total_tokens = self._message_token_count(self.system_prompt or "")
        for message in self._messages:
            if message.token_count is None:
                message.token_count = self._message_token_count(message.content or "")
            total_tokens += message.token_count
        self.current_usage = TokenUsage(total_tokens=total_tokens)

    def add_user_message(
        self,
        content: str,
        tool_calls: List[Dict[str, Any]] | None = None,
        images: List[ToolImage] | None = None,
    ) -> None:
        item = MessageItem(
            role=MessageRole.USER,
            content=content or "", 
            token_count=self._message_token_count(content or ""),
            tool_calls=tool_calls or [],
            images=images or [],
        )
        self._messages.append(item)
        self._refresh_current_usage()
 
    def add_assistant_message(self, content: str, tool_calls: List[Dict[str, Any]] | None = None) -> None:
        item = MessageItem(
            role=MessageRole.ASSISTANT,
            content=content or "", 
            token_count=self._message_token_count(content or ""),
            tool_calls=tool_calls or []
        )
        self._messages.append(item)
        self._refresh_current_usage()
 
    def get_messages(self) -> List[Dict[str, Any]]:
        messages: List[Dict[str, Any]] = []
        if self.system_prompt:
            messages.append({"role": "system", "content": self.system_prompt})
        for item in self._messages:
            messages.append(item.to_dict())
        return messages
    
    def add_tool_result(self, tool_result: ToolResultMessage) -> None:
        item = MessageItem(
            role=MessageRole.TOOL,
            content=tool_result.content or "",
            token_count=self._message_token_count(tool_result.content or ""),
            tool_call_id=tool_result.tool_call_id,
        )
        self._messages.append(item)
        self._refresh_current_usage()

    def set_usage(self, usage: TokenUsage) -> None:
        self.total_usage += usage
        self.latest_usage = usage
        self._refresh_current_usage()

    def current_context_tokens(self) -> int:
        self._refresh_current_usage()
        return self.current_usage.total_tokens

    def needs_compression(self) -> bool:
        context_limit = self.config.model.context_window
        if not context_limit:
            return False
        return self.current_context_tokens() > int(context_limit * 0.8)

    def replace_with_compressed_summary(self, summary: str, preserve_last_n: int = 2) -> None:
        preserved_messages = self._messages[-preserve_last_n:] if preserve_last_n > 0 else []
        self._messages = []
        self.add_user_message(
            "compressed summary of previous context:\n" + summary,
        )
        self.add_assistant_message(
            "understood. i will use this compressed summary together with the preserved recent context."
        )
        self._messages.extend(preserved_messages)
        self._refresh_current_usage()

   
    def clear(self) -> None:
        self._messages = []
        self.current_usage = TokenUsage()
        self.latest_usage = TokenUsage()
        self.total_usage = TokenUsage()




from __future__ import annotations

import asyncio

from pydantic import BaseModel, Field

from pegasus.agent.results import AgentEventType
from pegasus.config.config import Config
from pegasus.tools.base import Tool
from pegasus.tools.data import ToolInvocation, ToolResult, ToolType
from pegasus.utils.logger import logger


class SubAgentParams(BaseModel):
    goal: str = Field(..., description="The goal or instruction for the subagent to complete.")
    max_turns: int = Field(20, ge=1, le=100, description="The maximum number of turns for the subagent run.")
    timeout_seconds: int = Field(
        300,
        ge=10,
        le=1800,
        description="The maximum runtime in seconds for the subagent run.",
    )


class SubAgentTool(Tool):
    name: str = "subagent"
    description: str = (
        "Delegate a focused goal to a child agent. "
        "The subagent gets the caller's tools, runs its own agentic loop once, and returns its final answer."
    )
    type: ToolType = ToolType.SUB_AGENT
    schema: SubAgentParams = SubAgentParams

    def __init__(self, config: Config) -> None:
        self._config = config

    async def _run_subagent(self, params: SubAgentParams) -> ToolResult:
        from pegasus.agent.agent import Agent

        sub_config = Config(**self._config.model_dump(exclude_none=True))
        sub_config.max_turns = params.max_turns

        agent = Agent(sub_config)
        await agent.startup()

        final_response: str | None = None
        tool_names: list[str] = []
        try:
            async def _consume() -> None:
                nonlocal final_response
                async for event in agent.run(params.goal):
                    if event.type == AgentEventType.TEXT_COMPLETE:
                        final_response = event.data.get("content", "")
                        logger.info(f"subagent final response: {final_response}")
                    elif event.type == AgentEventType.TOOL_CALL_COMPLETE:
                        tool_name = event.data.get("name")
                        if tool_name:
                            tool_names.append(tool_name)
                            logger.info(
                                f"subagent tool call completed: {tool_name} success={event.data.get('success', False)}"
                            )

            await asyncio.wait_for(_consume(), timeout=params.timeout_seconds)
        except asyncio.TimeoutError:
            return ToolResult.error_result(
                f"subagent timed out after {params.timeout_seconds} seconds",
                metadata={"timeout_seconds": params.timeout_seconds},
            )
        finally:
            await agent.cleanup()

        output_lines = [
            "subagent completed successfully.",
            f"goal: {params.goal}",
            f"tools used: {', '.join(tool_names) if tool_names else 'none'}",
            "",
            "final response:",
            final_response if final_response else "no final response",
        ]
        return ToolResult.success_result(
            output="\n".join(output_lines),
            metadata={"tool_names": tool_names, "max_turns": params.max_turns},
        )

    async def _execute(self, invocation: ToolInvocation) -> ToolResult:
        params = SubAgentParams(**invocation.params)
        return await self._run_subagent(params)

    async def execute(self, invocation: ToolInvocation) -> ToolResult:
        try:
            return await self._execute(invocation)
        except Exception as e:
            return ToolResult.error_result(f"error executing subagent tool with error: {e}")



import asyncio
from typing import Dict

from pegasus.config.config import Config
from pegasus.mcp.client import MCPClient, MCPClientStatus
from pegasus.utils.logger import logger


class MCPManager:
    def __init__(self, config: Config):
        self._config = config
        self._clients: Dict[str, MCPClient] = {}
        self._initialized = False

    @property
    def initialized(self) -> bool:
        return self._initialized

    @property
    def connected_clients(self) -> Dict[str, MCPClient]:
        return {
            name: client
            for name, client in self._clients.items()
            if client.status == MCPClientStatus.CONNECTED
        }

    async def initialize(self) -> None:
        if self._initialized:
            return
        mcp_configs = self._config.mcp_servers
        if not mcp_configs:
            logger.info("[mcp manager] no mcp servers configured🤔")
            self._initialized = True
            return
        for name, server_config in mcp_configs.items():
            if not server_config.enabled:
                logger.info(f"[mcp manager] mcp server {name} disabled, skipping startup🚀")
                continue
            self._clients[name] = MCPClient(name, server_config, self._config.cwd)
        if not self._clients:
            logger.info("[mcp manager] no enabled mcp servers to start🚀")
            self._initialized = True
            return

        client_names = list(self._clients.keys())
        connect_tasks = [
            asyncio.wait_for(
                self._clients[name].connect(),
                timeout=mcp_configs[name].startup_timeout_sec,
            )
            for name in client_names
        ]
        results = await asyncio.gather(*connect_tasks, return_exceptions=True)
        connected_count = 0
        failed_count = 0
        for name, result in zip(client_names, results, strict=False):
            client = self._clients[name]
            if isinstance(result, Exception):
                failed_count += 1
                logger.error(f"[mcp manager] mcp server {name} failed to initialize💣: {result}")
                continue
            if client.status == MCPClientStatus.CONNECTED:
                connected_count += 1
                logger.info(f"[mcp manager] {name} mcp initialized with {len(client.tools)} tools🚀")
                continue
            failed_count += 1
            logger.warning(f"[mcp manager] mcp server {name} finished startup with status {client.status}💣")
        self._initialized = True
        summary_emoji = "💣" if failed_count else "🚀"
        logger.info(f"[mcp manager] startup complete: {connected_count} connected, {failed_count} failed{summary_emoji}")

    async def shutdown(self) -> None:
        await asyncio.gather(
            *[asyncio.wait_for(client.disconnect(), timeout=10.0) for client in self._clients.values()],
            return_exceptions=True,
        )
        self._clients.clear()
        self._initialized = False



import click
import asyncio
from pegasus.agent.agent import Agent
from pegasus.agent.results import AgentEventType
from pegasus.renderers.tui import TUI
import sys
from pathlib import Path
from pegasus.config.config import Config
from pegasus.config.loader import load_config


class CLI:
    def __init__(self, config: Config) -> None:
        self._config = config
        self.agent: Agent | None = None
        self.tui = TUI(self._config)

    async def startup(self) -> None:
        self.agent = Agent(self._config)
        await self.agent.startup()
        self.tui.print_welcome(title="Pegasus CLI", lines=[
            f"Model: {self._config.model_name}",
            f"CWD: {self._config.cwd}",
            "Commands: /help /config /model /exit"])

    async def run_single(self, message: str) -> None:
        await self._process_message(message)

    async def run_interactive(self) -> bool:
        while True:
            try:
                user_input = self.tui._console.input("\n[user]> [/user]").strip()

                if not user_input:
                    continue

                if user_input.startswith("/"):
                    should_continue = self._handle_command(user_input)
                    if not should_continue:
                        break
                else:
                    await self._process_message(user_input)
            except KeyboardInterrupt:
                self.tui._console.print("\n[dim] Use \\exit to quit[/dim]")
                break
            except EOFError:
                break
        self.tui._console.print("\n[dim] Goodbye Ol Friend![/dim]")
        return True

    def _handle_command(self, command: str) -> bool:
        command = command[1:].strip().lower()
        parts = command.split(maxsplit=1)
        cmd_name = parts[0]
        cmd_args = parts[1] if len(parts) > 1 else None
        if cmd_name == "exit" or cmd_name == "quit":
            return False
        elif cmd_name == "help":
            self.tui.print_help()
            return True
        elif cmd_name == "clear":
            self.agent.session.context_manager.clear()
            return True
        elif cmd_name == "config":
            self.tui.print_config()
            return True
        elif cmd_name == "model":
            if cmd_args:
                self._config.model_name = cmd_args
                self.tui.print_config()
                return True
            else:
                self.tui._console.print(f"[error]Model name is required[/error]")
                return True
    

    def _get_tool_kind(self, tool_name: str) -> str | None:
        tool_kind = None
        tool = self.agent.session.tool_registry.get(tool_name)
        if not tool:
            tool_kind = None
        tool_kind = tool.type
        return tool_kind

    async def _process_message(self, message: str) -> str | None:
        if not self.agent:
            return None
        assistant_streaming = False
        reasoning_streaming = False
        assistant_started = False
        reasoning_started = False
        final_response: str | None = None
        async for event in self.agent.run(message):
            if event.type == AgentEventType.TEXT_DELTA:
                content = event.data.get('content', '')
                if not assistant_streaming:
                    assistant_started = True
                    self.tui._begin_assistant()
                    assistant_streaming = True
                self.tui.stream_assistant_delta(content)
            elif event.type == AgentEventType.REASONING_DELTA:
                content = event.data.get('reasoning', '')
                if not reasoning_streaming:
                    reasoning_started = True
                    self.tui._begin_reasoning()
                    reasoning_streaming = True
                self.tui.stream_reasoning_delta(content)
            elif event.type == AgentEventType.REASONING_COMPLETE:
                if reasoning_streaming and reasoning_started:
                    reasoning_streaming = False
                    self.tui._end_reasoning()
            elif event.type == AgentEventType.TEXT_COMPLETE:
                if assistant_streaming and assistant_started:
                    assistant_streaming = False
                    self.tui._end_assistant()
            elif event.type == AgentEventType.AGENT_ERROR:
                error = event.data.get("error", "Unknown error")
                self.tui._console.print(f"[error]Error: {error}[/error]")
            elif event.type == AgentEventType.TOOL_CALL_START:
                if reasoning_streaming and reasoning_started:
                    reasoning_streaming = False
                    self.tui._end_reasoning()
                tool_name = event.data.get("name", "")
                call_id = event.data.get("call_id", "")
                arguments = event.data.get("arguments", {})
                tool_kind = self._get_tool_kind(tool_name)
                self.tui.tool_call_start(call_id, tool_name, tool_kind, arguments)
            elif event.type == AgentEventType.TOOL_CALL_COMPLETE:
                tool_name = event.data.get("name", "")
                call_id = event.data.get("call_id", "")
                tool_kind = self._get_tool_kind(tool_name)
                success = event.data.get("success", False)
                output = event.data.get("output", "")
                metadata = event.data.get("metadata", {})
                truncated = event.data.get("truncated", False)
                error = event.data.get("error", None)
                diff = event.data.get("diff", None)
                exit_code = event.data.get("exit_code", None)
                self.tui.tool_call_complete(call_id, tool_name, tool_kind, success, 
                                             output, error, metadata, truncated, diff, exit_code)
        return final_response

async def _run_cli(message: str | None = None, cwd: Path = Path.cwd()) -> None:
    try:
        config = load_config(cwd)
    except Exception as e:
        click.echo(f"Error loading config: {e}", err=True)
        sys.exit(1)
    cli = CLI(config)
    await cli.startup()
    if message:
        result = await cli.run_single(message)
    else:
        result = await cli.run_interactive()
    if result is None:
        click.echo("No response from agent", err=True)
        sys.exit(1)


@click.command()
@click.option('--message', type=str, help='The message to send to the chat completion.')
@click.option('--cwd', type=click.Path(exists=True, file_okay=False, path_type=Path), help='The current working directory.' , default=Path.cwd())
def run_cli(message: str | None = None, cwd: Path = Path.cwd()) -> None:
    """sync entry point that runs the async _run_cli."""
    asyncio.run(_run_cli(message=message, cwd=cwd))


from pathlib import Path


def resolve_path(base: str | Path, path: str | Path) -> Path:
    path = Path(path)
    if path.is_absolute():
        return path.resolve()
    return Path(base).resolve() / path

def ensure_parent_directory(path: str | Path) -> Path:
    path: Path = Path(path)
    path.parent.mkdir(parents=True, exist_ok=True)
    return path.resolve()

def display_path_rel_to_cwd(path: str, cwd: Path| None) -> str:
    try:
        p = Path(path)
    except Exception:
        return path
    if cwd:
        try:
            return str(p.relative_to(cwd))
        except ValueError:
            return str(p)
        except Exception:
            return str(p)
    return str(p)

def is_binary_file(path: str | Path) -> bool:
    try:
        with open(path, "rb") as file:
            chunk = file.read(8192)
            return b"\x00" in chunk or b"\xFF" in chunk
    except (OSError, IOError):
        return False


from typing import AsyncGenerator
from typing import List, Dict, Any
import json
from pegasus.agent.results import AgentEvent, AgentEventType
from pegasus.context.data import ToolResultMessage
from pegasus.client.response import TokenUsage, ToolCall, StreamEventType
from pegasus.session.session import Session
from pegasus.config.config import Config
from pegasus.utils.logger import logger
from uuid import uuid4


def _assistant_message_tool_calls(tool_calls: List[ToolCall]) -> List[Dict[str, Any]]:
    return [
        {
            "id": tc.call_id,
            "type": "function",
            "function": {"name": tc.name, "arguments": json.dumps(tc.arguments) if isinstance(tc.arguments, dict) else (tc.arguments or "{}")},
        }
        for tc in tool_calls
    ]

class Agent:
    def __init__(self, config: Config) -> None:
        self.id = str(uuid4())
        self._config = config
        self.session: Session | None = None

    def __str__(self) -> str:
        return f"Agent(config={self._config})"
    
    def __repr__(self) -> str:
        return self.__str__()

    async def run(self, message: str) -> AsyncGenerator[AgentEvent, None]:
        yield AgentEvent.agent_start(message)
        self.session.context_manager.add_user_message(message)
        final_response: str | None = None
        usage: TokenUsage | None = None
        async for event in self._agentic_loop():
            yield event
            if event.type == AgentEventType.TEXT_COMPLETE:
                final_response = event.data['content']
        yield AgentEvent.agent_end(final_response, usage)

    async def _agentic_loop(self) -> AsyncGenerator[AgentEvent, None]:
        # an agentic loop is a full execution cycle of a react agentic loop
        yield AgentEvent.loop_start()
        for turn in range(self._config.max_turns):
            # each turn is a single llm action
            await self.session.increment_turn()
            yield AgentEvent.turn_start()
            if self.session.context_manager.needs_compression():
                current_tokens = self.session.context_manager.current_context_tokens()
                context_limit = self._config.model.context_window or 0
                logger.info(
                    f"compressing context at {current_tokens} tokens with limit {context_limit}"
                )
                summary, compression_usage = await self.session.chat_compressor.compress(self.session.context_manager)
                if summary:
                    self.session.context_manager.replace_with_compressed_summary(summary)
                if compression_usage:
                    self.session.context_manager.set_usage(compression_usage)
            response_text = ''
            messages = self.session.context_manager.get_messages()
            tool_calls: List[ToolCall] = []
            usage: TokenUsage | None = None
            is_reasoning = False

            async for event in self.session.client.chat_completion(messages=messages, tools=self.session.tool_registry.get_schemas(), stream=True):
                if event.type == StreamEventType.TEXT_DELTA:
                    if event.text_delta and event.text_delta.content:
                        content = event.text_delta.content
                        response_text += content
                        if is_reasoning:
                            is_reasoning = False
                            yield AgentEvent.reasoning_complete()
                        yield AgentEvent.text_delta(content)
                    if event.reasoning_delta and event.reasoning_delta.reasoning:
                        reasoning = event.reasoning_delta.reasoning
                        is_reasoning = True
                        yield AgentEvent.reasoning_delta(reasoning)

                elif event.type == StreamEventType.TOOL_CALL_COMPLETE:
                    if event.tool_calls:
                        tool_calls.extend(event.tool_calls)
                elif event.type == StreamEventType.ERROR:
                    yield AgentEvent.agent_error(event.error if event.error else 'Unknown error', {})
                if event.type == StreamEventType.MESSAGE_COMPLETE:
                    usage = event.usage
            self.session.context_manager.add_assistant_message(response_text or "", tool_calls=_assistant_message_tool_calls(tool_calls))
            if response_text:
                yield AgentEvent.text_complete(response_text)
            
            if not tool_calls:
                if usage:
                    self.session.context_manager.set_usage(usage)
                # self.session.context_manager.prune_tool_outputs() # TODO: add context pruning later
                yield AgentEvent.turn_end()
                yield AgentEvent.loop_end()
                return

            tool_results: List[ToolResultMessage] = []
            for tool_call in tool_calls:
                yield AgentEvent.tool_call_start(tool_call.call_id, tool_call.name, tool_call.arguments)
                result = await self.session.tool_registry.invoke(tool_call.name, tool_call.arguments, self._config.cwd)
                tool_results.append(
                    ToolResultMessage(
                        tool_call_id=tool_call.call_id,
                        content=result.to_model_output(),
                        is_error=not result.success,
                        images=result.images,
                        metadata=result.metadata,
                    )
                )
                yield AgentEvent.tool_call_complete(
                    tool_call.call_id,
                    tool_call.name,
                    result.success,
                    result.output,
                    result.metadata,
                    result.truncated,
                    result.error,
                    result.diff,
                    result.exit_code,
                    result.images,
                )
            
            for tool_result in tool_results:
                self.session.context_manager.add_tool_result(tool_result)
            followup_images = [image for tool_result in tool_results for image in tool_result.images]
            if followup_images:
                followup_messages: list[str] = []
                for tool_call, tool_result in zip(tool_calls, tool_results, strict=False):
                    visual_context_message = tool_result.metadata.get("visual_context_message")
                    if visual_context_message:
                        followup_messages.append(f"{tool_call.name}: {visual_context_message}")
                followup_content = (
                    "\n".join(followup_messages)
                    if followup_messages
                    else "visual context from the previous tool call is attached. inspect these images before deciding the next action."
                )
                self.session.context_manager.add_user_message(
                    followup_content,
                    images=followup_images,
                )

            if usage:
                self.session.context_manager.set_usage(usage)            
            # self._session._context_manager.prune_tool_outputs() # TODO: add context pruning later
            yield AgentEvent.turn_end()
        yield AgentEvent.loop_end()
    
    async def startup(self) -> None:
        # initialize session if not already initialized
        logger.info(f"starting up agent")
        if self.session is None:
            self.session = Session(self._config)
            await self.session.initialize()

    async def cleanup(self) -> None:
        logger.info(f"cleaning up agent")
        if self.session is not None:
            await self.session.cleanup()
            self.session = None


if __name__ == "__main__":
    import asyncio

    async def main():
        config = Config()
        agent = Agent(config)
        await agent.startup()
        async for event in agent.run("Hello, how are you?"):
            logger.info(f"Event: {event}")
        await agent.cleanup()
        logger.info(f"Agent cleaned up")
    asyncio.run(main())


from pathlib import Path
import jinja2
from pegasus.config.config import Config
import json
import os, sys
from datetime import datetime
from pegasus.tools.base import Tool

PROMPTS_DIR = Path(__file__).resolve().parent
template_loader = jinja2.FileSystemLoader(searchpath=str(PROMPTS_DIR))
template_env = jinja2.Environment(loader=template_loader)

def get_system_prompt(config: Config) -> str:
    parts = []
    # identity and role
    parts.append(_get_identity_section(config))
    # environment
    parts.append(_get_environment_section(config))
    # security guidelines
    parts.append(_get_security_guidelines_section(config))
    # operational guidelines
    parts.append(_get_operational_section(config))
    # memory
    parts.append(_get_memory_section(config))
    return "\n\n".join(parts)

def _get_identity_section(config: Config) -> str:
    """
    retreives the identity and role of the agent.
    the identity allows the model to know which parts of its weights it should activate in order to perform the task properly
    """
    config
    template = template_env.get_template("identity.j2")
    return template.render()

def _get_shell_info() -> str:
    """get shell information based on platform."""
    if sys.platform == "darwin":
        return "/bin/zsh"
    elif sys.platform == "win32":
        return "PowerShell/cmd.exe"
    else:
        return "/bin/bash"

def _get_environment_section(config: Config) -> str:
    """
    retreives the environment section of the prompt.
    this determines the environment of the agent and how it should be used.
    """
    config
    template = template_env.get_template("environment.j2")
    cwd = config.cwd.as_posix()
    shell_info = _get_shell_info()
    return template.render({"now": datetime.now(), "os_info": os.uname().sysname, "cwd": cwd, "shell_info": shell_info})

def _get_security_guidelines_section(config: Config) -> str:
    """
    retreives the security guidelines section of the prompt.
    modify security guidelines to ensure the agent is secure and does not harm the user or the system.
    """
    template = template_env.get_template("security_guidelines.j2")
    return template.render()

def _get_operational_section(config: Config) -> str:
    """
    retreives the operational guidelines section of the prompt.
    this determines the tone and style of the agents responses.
    """
    config
    template = template_env.get_template("operational.j2")
    return template.render()

def _get_tools_section(config: Config, tools: list[Tool]) -> str:
    """
    retreives the tools section of the prompt.
    this determines the tools that the agent has access to.
    """
    regular_tools = [t for t in tools if not t.name.startswith("subagent_")]
    subagent_tools = [t for t in tools if t.name.startswith("subagent_")]

    guidelines = ""
    for tool in regular_tools:
        description = tool.description
        if len(description) > 100:
            description = description[:100] + "..."
        guidelines += f"- **{tool.name}**: {description}\n"

    if subagent_tools:
        guidelines += "\n## Sub-Agents\n\n"
        for tool in subagent_tools:
            description = tool.description
            if len(description) > 100:
                description = description[:100] + "..."
            guidelines += f"- **{tool.name}**: {description}\n"
    template = template_env.get_template("tools.j2")
    return template.render({"guidelines": guidelines, "subagent_tools": bool(subagent_tools)})

def _get_memory_section(config: Config) -> str:
    """
    retreives the memory section of the prompt.
    this determines the memory of the agent and how it should be used.
    """
    memory: Path = Path(os.path.join(config.cwd, "pegasus_memory.json"))
    if not memory.exists():
        return ""
    with open(memory, "r") as f:
        memory = json.load(f)
    template = template_env.get_template("memory.j2")
    return template.render({"memory": json.dumps(memory, indent=4)})


import tiktoken

def get_tokenizer(model: str):
    try:
        encoding = tiktoken.encoding_for_model(model)
        return encoding.encode
    except Exception:
        encoding = tiktoken.get_encoding("cl100k_base")
        return encoding.encode
    
def count_tokens(text: str, model: str) -> int:
    tokenizer = get_tokenizer(model)
    if tokenizer:
        return len(tokenizer(text))
    return estimate_tokens(text)

def estimate_tokens(text: str) -> int:
    return max(1, len(text) // 4)


def _truncate_by_chars(text: str, target_tokens: int, suffix: str, model: str) -> str:
    low, high = 0, len(text)
    while low < high:
        mid = (low + high) // 2
        if count_tokens(text[:mid], model) <= target_tokens:
            low = mid
        else:
            high = mid - 1

    return text[:low] + suffix.strip()


def _truncate_by_lines(text: str, target_tokens: int, suffix: str, model: str) -> str:
    lines = text.split("\n")
    result_lines: list[str] = []
    current_tokens = 0

    for line in lines:
        line_tokens = count_tokens(line + "\n", model)
        if current_tokens + line_tokens > target_tokens:
            break
        result_lines.append(line)
        current_tokens += line_tokens
    if not result_lines:
        return _truncate_by_chars(text, target_tokens, suffix, model)
    return "\n".join(result_lines) + "\n" + suffix.strip()

def truncate_text(text: str, model: str, max_tokens: int, 
                  suffix: str = "\n...[truncated]", 
                  preserve_lines: bool = True) -> str:
    token_count = count_tokens(text, model)
    if token_count <= max_tokens:
        return text
    suffix_tokens = count_tokens(suffix, model)
    target_tokens = max_tokens - suffix_tokens
    
    if target_tokens <= 0:
        return suffix.strip()
    
    if preserve_lines:
        return _truncate_by_lines(text, target_tokens, suffix, model)


from pydantic import BaseModel, Field, PrivateAttr, model_validator
from typing import List, Any, Dict
from pathlib import Path
from enum import Enum
from dotenv import load_dotenv
import os

from art import text2art

_GREEN = "\033[32m"

load_dotenv()

class ModelConfig(BaseModel):
    name: str = Field(default="moonshotai/kimi-k2.5")
    base_url: str = Field(default="https://openrouter.ai/api/v1")
    temperature: float = Field(default=1, ge=0, le=2.0)
    context_window: int | None = 262_144
    tool_support: bool = False
    vision_support: bool = False
    nitro_support: bool = False
    cost_per_million_input_tokens: float = 0.0
    cost_per_million_output_tokens: float = 0.0
    api_key: str = Field(default=os.getenv("OPENROUTER_API_KEY"))
    open_source: bool = True

    @property
    def cost_per_input_token(self) -> float:
        """calculate cost per input token."""
        return self.cost_per_million_input_tokens / 1_000_000 if self.cost_per_million_input_tokens else 0.0

    @property
    def cost_per_output_token(self) -> float:
        """calculate cost per output token."""
        return self.cost_per_million_output_tokens / 1_000_000 if self.cost_per_million_output_tokens else 0.0

    def __str__(self) -> str:
        art_title = f"{_GREEN}{text2art(self.name)}"
        msg = (
            f"{art_title}\n"
            f"  MODEL NAME         : {self.name}\n"
            f"  BASE URL            : {self.base_url}\n"
            f"  TEMPERATURE         : {self.temperature}\n"
            f"  CONTEXT WINDOW      : {self.context_window}\n"
            f"  TOOL SUPPORT        : {self.tool_support}\n"
            f"  VISION SUPPORT      : {self.vision_support}\n"
            f"  NITRO SUPPORT       : {self.nitro_support}\n"
            f"  OPEN SOURCE         : {self.open_source}\n"
            f"  COST/INPUT TOKEN    : ${self.cost_per_input_token:.4f}\n"
            f"  COST/OUTPUT TOKEN   : ${self.cost_per_output_token:.4f}\n"
            f"  COST/MILLION INPUT  : ${self.cost_per_million_input_tokens:.4f}\n"
            f"  COST/MILLION OUTPUT : ${self.cost_per_million_output_tokens:.4f}\n"
            f"  API KEY             : {self.api_key}\n"
        )
        return msg

    def __repr__(self) -> str:
        return self.__str__()

MODEL_OPTIONS: Dict[str, ModelConfig] = {
    # open source large models
    "kimi-k2.5": ModelConfig(name="moonshotai/kimi-k2.5", base_url="https://openrouter.ai/api/v1", temperature=1, context_window=262_144, 
                              tool_support=True, vision_support=True,
                              cost_per_million_input_tokens=0.45, cost_per_million_output_tokens=2.20, nitro_support=False, open_source=True),
    "glm5": ModelConfig(name="z-ai/glm-5", base_url="https://openrouter.ai/api/v1", temperature=1, context_window=202_752, tool_support=True, 
                        vision_support=False,
                        cost_per_million_input_tokens=0.8, cost_per_million_output_tokens=2.56, nitro_support=False, open_source=True),
    "qwen3.5-122b": ModelConfig(name="qwen/qwen3.5-122b-a10b", base_url="https://openrouter.ai/api/v1", temperature=1, 
                                cost_per_million_input_tokens=0.26, cost_per_million_output_tokens=2.08, nitro_support=False,
                                context_window=262_144, tool_support=True, vision_support=True, open_source=True),
    "minimax-m2.5": ModelConfig(name="minimax/minimax-m2.5", base_url="https://openrouter.ai/api/v1", temperature=1,
                                 context_window=196_608, tool_support=True, vision_support=True,
                                 cost_per_million_input_tokens=0.295, cost_per_million_output_tokens=1.2, nitro_support=False, open_source=True),
    "olmo3.1-32b-instruct": ModelConfig(name="allenai/olmo-3.1-32b-instruct", base_url="https://openrouter.ai/api/v1", 
                                        temperature=1, context_window=65_536, tool_support=True, vision_support=False,
                                        cost_per_million_input_tokens=0.20, cost_per_million_output_tokens=0.60, nitro_support=False, open_source=True),
    "nemotron-3-nano-30b": ModelConfig(name="nvidia/nemotron-3-nano-30b-a3b", base_url="https://openrouter.ai/api/v1",
                                       temperature=1, context_window=262_144, tool_support=True, vision_support=False,
                                       cost_per_million_input_tokens=0.05, cost_per_million_output_tokens=0.20, nitro_support=False, open_source=True),
    "devstral2": ModelConfig(name="mistralai/devstral-2512", base_url="https://openrouter.ai/api/v1",
                             temperature=1, context_window=262_144, tool_support=True, vision_support=True,
                             cost_per_million_input_tokens=0.25, cost_per_million_output_tokens=0.75, nitro_support=False, open_source=True),
    "mercury2": ModelConfig(name="inception/mercury-2", base_url="https://openrouter.ai/api/v1", temperature=1, 
                            context_window=128_000, tool_support=True, vision_support=False,
                            cost_per_million_input_tokens=0.25, cost_per_million_output_tokens=0.75, nitro_support=False, open_source=True),

    # closed models
    "gpt-5.4": ModelConfig(name="openai/gpt-5.4", base_url="https://openrouter.ai/api/v1", temperature=1, 
                           context_window=1_050_000, tool_support=True, vision_support=True, nitro_support=True,
                           cost_per_million_input_tokens=2.5, cost_per_million_output_tokens=15, open_source=False),
    "codex-5.3": ModelConfig(name="openai/gpt-5.3-codex", base_url="https://openrouter.ai/api/v1", temperature=1, context_window=400_000, 
                             tool_support=True, vision_support=True, nitro_support=True,
                             cost_per_million_input_tokens=1.75, cost_per_million_output_tokens=14, open_source=False),
    "claude-sonnet-4.6": ModelConfig(name="anthropic/claude-sonnet-4.6", base_url="https://openrouter.ai/api/v1", 
                                     temperature=1, context_window=1_000_000, tool_support=True, vision_support=True,
                                     nitro_support=True,
                                     cost_per_million_input_tokens=3, cost_per_million_output_tokens=15, open_source=False),
    "claude-opus-4.6": ModelConfig(name="anthropic/claude-opus-4.6", base_url="https://openrouter.ai/api/v1", 
                                   temperature=1, context_window=1_000_000, tool_support=True, vision_support=True,
                                   nitro_support=True,
                                   cost_per_million_input_tokens=5, cost_per_million_output_tokens=25, open_source=False),

    "gemini-3.1-flash-lite": ModelConfig(name="google/gemini-3.1-flash-lite-preview", base_url="https://openrouter.ai/api/v1", 
                                         temperature=1, context_window=1_048_576, 
                                         tool_support=True, vision_support=True, nitro_support=True,
                                         cost_per_million_input_tokens=0.25, cost_per_million_output_tokens=1.5, open_source=False),
}

def list_default_model_names() -> list:
    """Return a list of all available default model option names."""
    return list(MODEL_OPTIONS.keys())

def get_model_config(model_name: str) -> ModelConfig:
    """Get ModelConfig for the given model name."""
    config = MODEL_OPTIONS.get(model_name)
    if not config:
        raise ValueError(f"No model found for '{model_name}'. Available: {', '.join(MODEL_OPTIONS.keys())}")
    return config


def list_open_source_model_names() -> list:
    """Return model keys for all open source models."""
    return [key for key, cfg in MODEL_OPTIONS.items() if cfg.open_source]


def list_closed_source_model_names() -> list:
    """Return model keys for all closed source models."""
    return [key for key, cfg in MODEL_OPTIONS.items() if not cfg.open_source]


def get_model_options_by_context_window() -> List[Dict[str, Any]]:
    """return model options as a list of dicts with model_key and context_window for sorting/filtering."""
    return [
        {"model_key": key, "context_window": cfg.context_window or 0}
        for key, cfg in MODEL_OPTIONS.items()
    ]


def get_model_options_by_input_token_cost() -> List[Dict[str, Any]]:
    """return model options as a list of dicts with model_key and cost_per_million_input_tokens for sorting/filtering."""
    return [
        {"model_key": key, "cost_per_million_input_tokens": cfg.cost_per_million_input_tokens}
        for key, cfg in MODEL_OPTIONS.items()
    ]


def get_model_options_by_output_token_cost() -> List[Dict[str, Any]]:
    """return model options as a list of dicts with model_key and cost_per_million_output_tokens for sorting/filtering."""
    return [
        {"model_key": key, "cost_per_million_output_tokens": cfg.cost_per_million_output_tokens}
        for key, cfg in MODEL_OPTIONS.items()
    ]


class ShellEnvironmentPolicy(BaseModel):
    ignore_default_excludes: bool = False
    exclude_patterns: List[str] = Field(default_factory=lambda: ["*KEY*", "*TOKEN*", "*SECRET*"])
    set_vars: dict[str, str] = Field(default_factory=dict)

class MCPServerConfig(BaseModel):
    enabled: bool = True
    startup_timeout_sec: float = 10.0

    # stdout transport
    command: str = Field(default="")
    args: List[str] = Field(default_factory=list)
    env: dict[str, str] = Field(default_factory=dict)
    cwd: Path | None = None

    # http/sse transport
    url: str | None = None

    @model_validator(mode="after")
    def validate_transport(self) -> 'MCPServerConfig':
        has_command = bool(self.command)
        has_url = bool(self.url)
        if not has_command and not has_url:
            raise ValueError("Either command or url must be provided")
        if has_command and has_url:
            raise ValueError("Only one of command or url must be provided")
        if has_command and not self.cwd:
            raise ValueError("cwd must be provided when command is provided")
        return self

class Config(BaseModel):
    model: ModelConfig = Field(default_factory=ModelConfig)
    cwd: Path = Field(default_factory=Path.cwd)
    max_turns: int = 100
    shell_environment: ShellEnvironmentPolicy = Field(default_factory=ShellEnvironmentPolicy)
    max_tool_output_tokens: int = 50_000
    mcp_servers: Dict[str, MCPServerConfig] = Field(default_factory=dict)
    _mcp_manager: Any | None = PrivateAttr(default=None)

    @property
    def api_key(self) -> str:
        return os.getenv('API_KEY', '')
    
    @property
    def base_url(self) -> str:
        return self.model.base_url
    
    @property
    def model_name(self) -> str:
        return self.model.name

    @model_name.setter
    def model_name(self, value: str) -> None:
        self.model.name = value

    @property
    def temperature(self) -> None:
        return self.model.temperature

    @temperature.setter
    def temperature(self, value: float) -> None:
        self.model.temperature = value

    def validate(self) -> List[str] | None:
        errors: List[str] = []
        if not self.api_key:
            errors.append("API key is required export API_KEY=sk-xxx from (https://openrouter.ai/keys) or set it in the .env file")

        if not self.cwd.exists():
            errors.append(f"Current working directory does not exist: {self.cwd}")
        
        if errors:
            return errors
        return None

    @property
    def mcp_manager(self) -> Any | None:
        return self._mcp_manager

    @mcp_manager.setter
    def mcp_manager(self, value: Any | None) -> None:
        self._mcp_manager = value


from pegasus.tools.base import Tool
from pegasus.tools.data import ToolType, ToolInvocation, ToolResult, FileDiff
from pegasus.config.config import Config
from pydantic import BaseModel, Field
from pegasus.utils.paths import resolve_path, ensure_parent_directory
from pathlib import Path


class EditFileParams(BaseModel):
    path: str = Field(..., description="The path to the file to edit")
    old_string: str = Field(..., description="The exact text to find and replace Must match exactly (including whitespace and identation) and must be unique in the file unless replace_all is true")
    new_string: str = Field(..., description="The text to replace old string with. Can be empty to delete the old string.")
    replace_all: bool = Field(False, description="Whether to replace all occurrences of the old_string or just the first one")

class EditFileTool(Tool):
    name: str = "edit_file"
    description: str = "Edit a file by replacing text. The `old_string` must match exactly "\
    "(including whitespace and identation) and must be unique in the file" \
    "unless replace_all is true. Use this for precise, surgical edits. "\
    "For creating new files or complete rewrites, use write_file instead."
    type: ToolType = ToolType.WRITE
    schema: EditFileParams = EditFileParams


    def __init__(self, config: Config) -> None:
        self._config = config


    def _no_match_error(self, old_string:str, content:str, path: Path) -> str:
        lines =  content.splitlines()
        partial_matches = []
        search_terms = old_string.splitlines()[:5]

        if search_terms:
            first_term = search_terms[0]
            for i, line in enumerate(lines, 1):
                if first_term in line:
                    partial_matches.append((i, line.strip()[:80]))
                    if len(partial_matches) >= 3:
                        break
        error_msg = f"old_string not found in file: {path.as_posix()}"
        if partial_matches:
            error_msg += f"\n\nPossible similar lines:\n"
            for line_num, line_preview in partial_matches:
                error_msg += f"\n Line {line_num}: {line_preview}"
            error_msg += f"\n\nMake sure old_string matches exactly (including whitespace and indentation)."
        else:
            error_msg += f"\n\nMake sure old_string matches exactly (including whitespace, indentation, line breaks and any invisible characters).Try reading the file again."
        return ToolResult.error_result(error_msg)

    async def _execute(self, invocation: ToolInvocation) -> ToolResult:
        params = EditFileParams(**invocation.params)
        path = resolve_path(invocation.cwd, params.path)
        if not path.exists() and not params.old_string:
            return ToolResult.error_result(f"File does not exist: {path.as_posix()}")
        file_exists = path.exists()
        if not file_exists:
            # TODO: the edit file tool should not be able to create a new file with the edit_file tool, we will remove this later
            path: Path = ensure_parent_directory(path)
            path.touch()
            path.write_text(params.new_string, encoding="utf-8")
            line_count = len(params.new_string.splitlines())
            byte_count = len(params.new_string.encode("utf-8"))
            return ToolResult.success_result(f"Created {path} {line_count} lines", 
                                             diff = FileDiff(path=path, old_content="", new_content=params.new_string, is_new_file=True, is_deletion=False), 
                                             metadata=dict(path=str(path.as_posix()), is_new_file=True, lines=line_count, bytes=byte_count))
        
        old_content = path.read_text(encoding="utf-8")
        if not old_content:
            return ToolResult.error_result(f"File is empty: {path.as_posix()} use write_file to create a new file")
        occurence_count = old_content.count(params.old_string)
        if occurence_count == 0:
            return self._no_match_error(params.old_string, old_content, path)
        
        if occurence_count > 1 and not params.replace_all:
            return ToolResult.error_result(f"old_string found {occurence_count} times in file: {path.as_posix()}. Use replace_all to replace all occurrences. Provide more context o make the match unique.", metadata=dict(path=str(path.as_posix()), occurence_count=occurence_count))
        
        if params.replace_all:
            new_content = old_content.replace(params.old_string, params.new_string)
            replaced_count = occurence_count
        else:
            new_content = old_content.replace(params.old_string, params.new_string, 1)
            replaced_count = 1
        if new_content == old_content:
            return ToolResult.error_result(f"No changes made to file: {path.as_posix()} old string is the same as new string. Check old_string and new_string.")
        path.write_text(new_content, encoding="utf-8")
        new_line_count = len(new_content.splitlines())
        old_line_count = len(old_content.splitlines())
        line_diff = new_line_count - old_line_count
        diff_message = ""
        if line_diff > 0:
            diff_message += f"+{line_diff} lines\n"
        elif line_diff < 0:
            diff_message += f"-{line_diff} lines\n"
        diff = FileDiff(path=path, old_content=old_content, new_content=new_content, is_new_file=file_exists, is_deletion=False)
        return ToolResult.success_result(output=f"Edited {path.as_posix()} {new_line_count} lines. {diff_message}", metadata=dict(path=str(path.as_posix()), is_new_file=file_exists, 
                                                                                                              lines=new_line_count,
                                                                                                              replaced_count=replaced_count), 
                                                                                                              diff=diff)
    

    async def execute(self, invocation: ToolInvocation) -> ToolResult:
        try:
            return await self._execute(invocation)
        except Exception as e:
            return ToolResult.error_result(f"Error executing edit_file tool with error: {e}")   


from pegasus.tools.base import Tool
from pegasus.tools.data import ToolType, ToolInvocation, ToolResult, FileDiff
from pegasus.config.config import Config
from pydantic import BaseModel, Field
from pegasus.utils.paths import resolve_path, ensure_parent_directory

class WriteFileParams(BaseModel):
    path: str = Field(..., description="The path to the file to write to (related to the working directory or absolute path)")
    create_dirs: bool = Field(True, description="Whether to create the parent directories if they do not exist")
    content: str = Field(..., description="The contents to write to the file")

class WriteFileTool(Tool):
    name: str = "write_file"
    description: str = "Write the contents of a text file. Create the file if it does not exist, "\
                        "overwrite the file if it exists. Parent directories are created automatically."\
                        "Use this for creating new files or completely replacing file contents"\
                        "For partial modifications, use the edit_file tool instead"
    type: ToolType = ToolType.WRITE
    schema: WriteFileParams = WriteFileParams

    def __init__(self, config: Config) -> None:
        self._config = config

    async def _execute(self, invocation: ToolInvocation) -> ToolResult:
        params = WriteFileParams(**invocation.params)
        path = resolve_path(invocation.cwd, params.path)
        is_new_file = not path.exists()
        old_content = None
        if not is_new_file:
            try:
                old_content = path.read_text(encoding="utf-8")
            except Exception as e:
                pass
        
        if is_new_file:
            if params.create_dirs:
                ensure_parent_directory(path)
            elif not path.parent.exists():
                return ToolResult.error_result(f"Parent directory does not exist: {path.parent.as_posix()}")
            path.touch()
        try:
            path.write_text(params.content)
        except Exception as e:
            return ToolResult.error_result(f"Error writing file with error: {e}")
        action = "created" if is_new_file else "overwrote"
        line_count = len(params.content.splitlines())
        diff = FileDiff(path=path, old_content=old_content, new_content=params.content, is_new_file=is_new_file, is_deletion=False)
        return ToolResult.success_result(output=f"{action} {path} {line_count} lines", metadata=dict(path=str(path.as_posix()), is_new_file=is_new_file, lines=line_count, bytes=len(params.content.encode("utf-8"))), diff=diff)
    
    async def execute(self, invocation: ToolInvocation) -> ToolResult:
        try:
            return await self._execute(invocation)
        except Exception as e:
            return ToolResult.error_result(f"Error writing file with error: {e}")


from __future__ import annotations

import base64
import mimetypes
from pathlib import Path
from pegasus.tools.data import ToolImage
from pegasus.utils.paths import ensure_parent_directory


def infer_image_mime_type(path: str | Path, fallback: str = "image/png") -> str:
    mime_type, _ = mimetypes.guess_type(str(path))
    if mime_type and mime_type.startswith("image/"):
        return mime_type
    return fallback


def image_file_to_data_url(path: str | Path, mime_type: str | None = None) -> str:
    path = Path(path)
    resolved_mime_type = mime_type or infer_image_mime_type(path)
    encoded = base64.b64encode(path.read_bytes()).decode("utf-8")
    return f"data:{resolved_mime_type};base64,{encoded}"


def tool_image_from_path(path: str | Path, mime_type: str | None = None) -> ToolImage:
    path = Path(path).resolve()
    resolved_mime_type = mime_type or infer_image_mime_type(path)
    return ToolImage(
        data_url=image_file_to_data_url(path, resolved_mime_type),
        mime_type=resolved_mime_type,
        path=path.as_posix(),
    )


def write_image_bytes(path: str | Path, content: bytes) -> Path:
    path = ensure_parent_directory(path)
    path.write_bytes(content)
    return path


def write_cv2_frame(path: str | Path, frame, params: list[int] | None = None) -> Path:
    import cv2

    path = ensure_parent_directory(path)
    if not cv2.imwrite(path.as_posix(), frame, params or []):
        raise RuntimeError(f"failed to write image to {path.as_posix()}")
    return path


def show_image(path: str | Path, window_name: str = "pegasus") -> None:
    import cv2

    image = cv2.imread(Path(path).as_posix())
    if image is None:
        return
    cv2.imshow(window_name, image)
    cv2.waitKey(1)



from pegasus.mcp.client import MCPClient, MCPClientStatus, MCPToolInfo
from pegasus.mcp.manager import MCPManager

__all__ = [
    "MCPClient",
    "MCPClientStatus",
    "MCPToolInfo",
    "MCPManager",
]



from pydantic import BaseModel, Field
from typing import Any
from pegasus.tools.base import Tool
from pegasus.tools.data import ToolType, ToolInvocation, ToolResult
from pegasus.utils.paths import resolve_path, is_binary_file
from pegasus.utils.text import estimate_tokens, truncate_text
from pegasus.config.config import Config
from pegasus.utils.logger import logger

MAX_FILE_SIZE = 1024 * 1024 * 10 # 10MB

class ReadFileParams(BaseModel):
    path: str = Field(..., description="The path to the file to read (related to the working directory or absolute path)")
    offset: int = Field(1, ge=1, description="The offset to read from the file")
    limit: int | None = Field(None, ge=1, description="The maximum number of bytes to read from the file")


class ReadFileTool(Tool):
    name: str = "read_file"
    description: str = "Read the contents of a text file. Returns the file content wiht the line numbers"\
                        "For large files, use offset and limit to read specific portions" \
                        "Cannot read binary files (images, executables, etc.)"
    type: ToolType = ToolType.READ
    schema: ReadFileParams = ReadFileParams

    def __init__(self, config: Config) -> None:
        self._config = config
        self._max_output_tokens = config.max_tool_output_tokens

    async def _execute(self, invocation: ToolInvocation) -> ToolResult:
        params = ReadFileParams(**invocation.params)
        path = resolve_path(invocation.cwd, params.path)
        if not path.exists():
            return ToolResult.error_result(f"File not found: {path}")
        
        if not path.is_file():
            return ToolResult.error_result(f"Path is not a file: {path}")
        
        # handle management of large files
        file_size = path.stat().st_size
        if file_size > MAX_FILE_SIZE:
            return ToolResult.error_result(f"File is too large to read: {file_size / (1024 * 1024): .2f} MB. Maximum allowed size is {MAX_FILE_SIZE / (1024 * 1024): .2f} MB")
        
        if is_binary_file(path):
            return ToolResult.error_result(f"Cannot read binary files: {path}")
        
        try:
            content = path.read_text(encoding="utf-8")
        except UnicodeDecodeError:
            content = path.read_text(encoding="latin-1")

        lines = content.splitlines()
        total_lines = len(lines)
        if total_lines == 0:
            return ToolResult.success_result(output="File is empty", metadata=dict(total_lines=0), truncated=False)
        
        start_idx = max(0, params.offset - 1)
        if params.limit is not None:
            end_idx = min(start_idx + params.limit, total_lines)
        else:
            end_idx = total_lines
        
        selected_lines = lines[start_idx:end_idx]
        formated_lines = [f"{i+start_idx+1: >6} | {line}" for i, line in enumerate(selected_lines)]

        output = "\n".join(formated_lines)
        token_count = estimate_tokens(output)
        truncated = False

        if token_count > self._max_output_tokens:
           output = truncate_text(output, self._config.model_name, self._max_output_tokens)
           truncated = True

        metadata_lines = []
        if start_idx > 0 or end_idx < total_lines:
            metadata_lines.append(
                f"Showing {end_idx - start_idx} of {total_lines} lines"
            )
        if metadata_lines:
            header = " | ".join(metadata_lines) + "\n\n"
            output = header + output
        
        return ToolResult.success_result(
            output=output,
            truncated = truncated,
            metadata=dict(path=path.as_posix(), total_lines=total_lines, shown_start=start_idx + 1, shown_end=end_idx)
        )

    async def execute(self, invocation: ToolInvocation) -> ToolResult:
        try:
            result = await self._execute(invocation)
            return result
        except Exception as e:
            return ToolResult.error_result(f"Error reading file with error: {e}")


from pegasus.config.config import Config
from pegasus.session.session import Session
from pegasus.config.loader import get_config_dir
import os
from datetime import datetime
from typing import Any
import json

class SessionSnapshot:
    session_id: str
    created_at: datetime
    updated_at: datetime
    turn_count: int
    messages: list[dict[str, Any]]

    def to_dict(self) -> dict[str, Any]:
        return {
            "session_id": self.session_id,
            "created_at": self.created_at.isoformat(),
            "updated_at": self.updated_at.isoformat(),
            "turn_count": self.turn_count,
            "messages": self.messages,
        }
    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> 'SessionSnapshot':
        return SessionSnapshot(
            session_id=data["session_id"],
            created_at=datetime.fromisoformat(data["created_at"]),
            updated_at=datetime.fromisoformat(data["updated_at"]),
            turn_count=data["turn_count"],
            messages=data["messages"],
        )

class PersistenceManager:
    def __init__(self, config: Config) -> None:
        self._config = config
        self._data_dir = get_config_dir()
        self._session_dir = self._data_dir / "sessions"
        self._session_dir.mkdir(parents=True, exist_ok=True)
        self._checkpoint_dir = self._data_dir / "checkpoints"
        self._checkpoint_dir.mkdir(parents=True, exist_ok=True)
        os.chmod(self._session_dir, 0o700)
        os.chmod(self._checkpoint_dir, 0o700)


    def save(self, snapshot: SessionSnapshot) -> None:
        session_file = self._session_dir / f"{snapshot.session_id}.json"
        with open(session_file, "w") as f:
            json.dump(snapshot.to_dict(), f, indent=4)
        os.chmod(session_file, 0o600)

    def load(self, session_id: str) -> SessionSnapshot | None:
        session_file = self._session_dir / f"{session_id}.json"
        if not session_file.exists():
            return None
        with open(session_file, "r") as f:
            data = json.load(f)
            return SessionSnapshot.from_dict(data)

    def list_sessions(self) -> list[SessionSnapshot]:
        session_files = list(self._session_dir.glob("*.json"))
        sessions = []
        for session_file in session_files:
            with open(session_file, "r") as f:
                data = json.load(f)
                sessions.append({
                    "session_id": data["session_id"],
                    "created_at": datetime.fromisoformat(data["created_at"]),
                    "updated_at": datetime.fromisoformat(data["updated_at"]),
                    "turn_count": data["turn_count"],
                })
        sessions = sorted(sessions, key=lambda x: x["updated_at"], reverse=True)
        return sessions
                

    def save_checkpoint(self, snapshot: SessionSnapshot, ) -> None:
        timestamp = datetime.now().isoformat()
        checkpoint_file = self._checkpoint_dir / f"{snapshot.session_id}_{timestamp}.json"
        with open(checkpoint_file, "w") as f:
            json.dump(snapshot.to_dict(), f, indent=4)
        os.chmod(checkpoint_file, 0o600)

    def load_checkpoint(self, session_id: str, timestamp: str) -> SessionSnapshot | None:
        checkpoint_file = self._checkpoint_dir / f"{session_id}_{timestamp}.json"
        if not checkpoint_file.exists():
            return None
        with open(checkpoint_file, "r") as f:
            data = json.load(f)
            return SessionSnapshot.from_dict(data)


from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Any, Type
from pydantic import BaseModel, ValidationError
from pegasus.tools.data import ToolInvocation, ToolResult, ToolType

class Tool(ABC):
    name: str = "base_tool"
    description: str = "A base tool class"
    type: ToolType = ToolType.WRITE

    @property
    def schema(self) -> dict[str, Any] | Type[BaseModel]:
        raise NotImplementedError("Every tool must implement the schema method")

    @abstractmethod
    def execute(self, invocation: ToolInvocation, *args, **kwargs) -> ToolResult:
        raise NotImplementedError("Every tool must implement the execute method")

    def validate_params(self, params: dict[str, Any]) -> list[str] | None:
        schema = self.schema
        missing_params = []
        if isinstance(schema, dict):
            for key in schema.keys():
                if key not in params:
                    missing_params.append(f"Missing required parameter: {key}")
            return missing_params
        elif isinstance(schema, type) and issubclass(schema, BaseModel):
            try:
                schema.model_validate(params).model_dump()
            except ValidationError as e:
                errors = []
                for error in e.errors():
                    field = ".".join(str(x) for x in error.get("loc", []))
                    msg = error.get("msg", "Validation error")
                    errors.append(f"Parameter {field}: {msg}")
                return errors
            except Exception as e:
                return [str(e)]
        return None
    
    def is_mutating(self, params: dict[str, Any]) -> bool:
        return self.type in [ToolType.WRITE, ToolType.BASH, ToolType.NETWORK, ToolType.MEMORY]
    
    def to_openai_schema(self) -> dict[str, Any]:
        schema = self.schema

        if isinstance(schema, type) and issubclass(schema, BaseModel):
            json_schema = schema.model_json_schema()
            result = {
                "name": self.name,
                "description": self.description,
                "parameters": {
                    "type": "object",
                    "properties": json_schema.get("properties", {}),
                    "required": json_schema.get("required", []),
                    "additionalProperties": False
                }
            }
            return result
        
        elif isinstance(schema, dict):
            result =  {
                "name": self.name,
                "description": self.description
            }
            if "parameters" in schema:
                result["parameters"] = schema["parameters"]
            else:
                result["parameters"] = schema
            return result
        
        raise ValueError(f"Invalid schema type for tool {self.name}: {type(schema)}")

    async def close(self) -> None:
        return None



from typing import Any

class AgentError(Exception):
    def __init__(self, message: str, *, details: dict[str, Any] | None = None, 
                 cause: Exception | None = None) -> None:
        super().__init__(message)
        self.message = message
        self.details = details or {}
        self.cause = cause

    def __str__(self) -> str:
        base = self.message
        if self.details:
            details_str = "\n".join([f"  {k}={v}" for k, v in self.details.items()])
            base = f"{base} ({details_str})"
        if self.cause:
            base = f"{base} [caused by: {self.cause}]"
        return base
    
    def to_dict(self) -> dict[str, Any]:
        return {
            "message": self.message,
            "details": self.details,
            "cause": self.cause
        }
    
class ConfigError(AgentError):
    def __init__(
            self, 
            message: str, 
            *, 
            config_key: str | None = None, 
            config_file: str | None = None,
            **kwargs: Any
    ) -> None:
        details = kwargs.pop("details", {}) or {}
        if config_key:
            details["config_key"] = config_key
        if config_file:
            details["config_file"] = config_file
        super().__init__(message, details=details, **kwargs)
        self.config_key = config_key
        self.config_file = config_file


from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, Any, List
from datetime import datetime
from pegasus.tools.data import ToolImage

class MessageRole(str, Enum):
    USER = "user"
    ASSISTANT = "assistant"
    SYSTEM = "system"
    TOOL = "tool"

@dataclass
class MessageItem:
    role: MessageRole
    content: str
    images: List[ToolImage] = field(default_factory=list)
    tool_call_id: str | None = None
    tool_calls: List[Dict[str, Any]] = field(default_factory=list)
    token_count: int | None = None
    pruned_at: datetime | None = None

    def to_dict(self) -> Dict[str, Any]:
        result: Dict[str, Any] = {"role": self.role.value}
        if self.tool_call_id:
            result["tool_call_id"] = self.tool_call_id
        if self.tool_calls:
            result["tool_calls"] = self.tool_calls
        content: str = self.content if self.content else ""
        if len(self.images) > 0:
            content: list[dict[str, Any]] = [{"type": "text", "text": content}] if content else []
            for image in self.images:
                content.append(image.to_content_part())
        result["content"] = content
        return result

@dataclass
class ToolResultMessage:
    tool_call_id: str
    content: str
    is_error: bool = False
    images: list[ToolImage] = field(default_factory=list)
    metadata: Dict[str, Any] = field(default_factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        content: str = self.content if self.content else ""
        if len(self.images) > 0:
            content: list[dict[str, Any]] = [{"type": "text", "text": content}] if content else []
            for image in self.images:
                content.append(image.to_content_part())
        result: Dict[str, Any] = {"role": "tool", "tool_call_id": self.tool_call_id, "content": content}
        return result


from pegasus.client.llm_client import LLMClient
from pegasus.client.response import TokenUsage, StreamEventType
from pegasus.context.manager import ContextManager
from pegasus.utils.logger import logger
from typing import Any

class ChatCompressor:
    def __init__(self, client: LLMClient) -> None:
        self._client = client

    def _render_content(self, content: Any) -> str:
        if isinstance(content, str):
            return content
        if isinstance(content, list):
            parts: list[str] = []
            for item in content:
                if not isinstance(item, dict):
                    parts.append(str(item))
                    continue
                if item.get("type") == "text":
                    parts.append(item.get("text", ""))
            return " ".join(part for part in parts if part).strip()
        return str(content)

    def _format_history_for_compaction(self, messages: list[dict[str, Any]]) -> str:
        lines: list[str] = []
        for message in messages:
            rendered_content = self._render_content(message.get("content", ""))
            if not rendered_content:
                continue
            tool_call_id = message.get("tool_call_id")
            prefix = message.get("role", "unknown")
            if tool_call_id:
                prefix = f"{prefix}({tool_call_id})"
            lines.append(f"{prefix}: {rendered_content}")
        return "\n".join(lines)
    
    async def compress(self, context_manager: ContextManager) -> tuple[ str | None, TokenUsage | None]:
        messages = context_manager.get_messages()[1:]

        if len(messages) < 3:
            return None, None
        
        compression_message = [
            {
                "role": "system",
                "content": (
                    "You compress long assistant conversations into a concise working summary. "
                    "Preserve trajectory, semantics, and instructional information with high fidelity. "
                    "Keep user goals, task progress, key constraints, decisions, important facts, relevant tool outputs, "
                    "open questions, pending work, and anything the agent must remember to continue correctly."
                ),
            },
            {
                "role": "user",
                "content": (
                    "compress the following conversation history into a compact but high-signal summary "
                    "that can replace older context while preserving continuity. "
                    "the summary must retain: "
                    "1. trajectory: what has already been tried, discovered, changed, or decided. "
                    "2. semantics: the actual meaning of prior observations, results, errors, and tool outputs. "
                    "3. instructional information: user requirements, constraints, preferences, and explicit do or do not rules.\n\n"
                    + self._format_history_for_compaction(messages)
                ),
            }
        ]
        
        try:
            summary = ""
            usage = None
            async for event in self._client.chat_completion(messages=compression_message, stream=False):
                if event.type == StreamEventType.MESSAGE_COMPLETE:
                    usage = event.usage
                    if event.text_delta and event.text_delta.content:
                        summary += event.text_delta.content
            summary = summary.strip()
            if not summary or not usage:
                logger.error("failed to compress context: missing summary or usage")
                return None, None
            return summary, usage
        except Exception as e:
            logger.error(f"failed to compress context: {e}")
            return None, None



from __future__ import annotations

import time
from datetime import datetime
from pathlib import Path
from typing import Any

from pydantic import BaseModel, Field

from pegasus.config.config import Config
from pegasus.tools.base import Tool
from pegasus.tools.data import ToolInvocation, ToolResult, ToolType
from pegasus.utils.images import tool_image_from_path, write_cv2_frame


class VisionCaptureParams(BaseModel):
    duration_seconds: int = Field(
        ...,
        ge=10,
        le=50,
        description="The capture duration in seconds. Must be between 10 and 50 seconds.",
    )
    camera_index: int = Field(0, ge=0, description="The camera device index to capture from.")
    max_frames: int = Field(
        6,
        ge=2,
        le=10,
        description="The maximum number of frames to keep for the next model turn.",
    )


class VisionCaptureTool(Tool):
    name: str = "vision_capture"
    description: str = (
        "Capture a short window of camera frames and return representative images "
        "for visual follow-up in the next model step."
    )
    type: ToolType = ToolType.READ
    schema: VisionCaptureParams = VisionCaptureParams

    def __init__(self, config: Config) -> None:
        self._config = config

    def _artifact_dir(self) -> Path:
        stamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        return self._config.cwd / ".pegasus" / "artifacts" / "vision_capture" / stamp

    def _capture_frames(self, params: VisionCaptureParams) -> dict[str, Any]:
        import cv2

        window_name = f"pegasus - {params.duration_seconds}s"
        window_created = False
        cap = cv2.VideoCapture(params.camera_index)
        if not cap.isOpened():
            raise RuntimeError(f"unable to open camera index {params.camera_index}")

        try:
            cv2.namedWindow(window_name, cv2.WINDOW_NORMAL)
            window_created = True
            start = time.monotonic()
            sample_interval = max(0.5, params.duration_seconds / params.max_frames)
            next_sample = start
            captured_paths: list[str] = []
            frame_count = 0
            artifact_dir = self._artifact_dir()

            while True:
                ret, frame = cap.read()
                if not ret:
                    raise RuntimeError("failed to read frame from camera")

                frame_count += 1
                display_frame = cv2.resize(frame, (800, int(frame.shape[0] * 800 / frame.shape[1])))
                elapsed_seconds = max(0.0, time.monotonic() - start)
                remaining_seconds = max(0.0, params.duration_seconds - elapsed_seconds)
                cv2.rectangle(display_frame, (0, 0), (display_frame.shape[1], 48), (0, 0, 0), -1)
                cv2.putText(
                    display_frame,
                    f"capture duration: {params.duration_seconds}s",
                    (16, 20),
                    cv2.FONT_HERSHEY_SIMPLEX,
                    0.6,
                    (255, 255, 255),
                    2,
                    cv2.LINE_AA,
                )
                cv2.putText(
                    display_frame,
                    f"time remaining: {remaining_seconds:.1f}s",
                    (16, 40),
                    cv2.FONT_HERSHEY_SIMPLEX,
                    0.55,
                    (255, 255, 255),
                    1,
                    cv2.LINE_AA,
                )
                cv2.imshow(window_name, display_frame)
                cv2.waitKey(1)

                now = time.monotonic()
                if now >= next_sample and len(captured_paths) < params.max_frames:
                    frame_index = len(captured_paths) + 1
                    frame_path = write_cv2_frame(
                        artifact_dir / f"frame_{frame_index:02d}.jpg",
                        frame,
                        [int(cv2.IMWRITE_JPEG_QUALITY), 90],
                    )
                    captured_paths.append(frame_path.as_posix())
                    next_sample = now + sample_interval

                if now - start >= params.duration_seconds:
                    break

                time.sleep(0.03)

            return {
                "paths": captured_paths,
                "frame_count": frame_count,
            }
        finally:
            cap.release()
            if window_created:
                try:
                    cv2.destroyWindow(window_name)
                    cv2.waitKey(1)
                except Exception:
                    # fall back to a broader cleanup if the named window is already gone
                    cv2.destroyAllWindows()
                    cv2.waitKey(1)

    async def _execute(self, invocation: ToolInvocation) -> ToolResult:
        params = VisionCaptureParams(**invocation.params)
        capture_result = self._capture_frames(params)
        image_paths = capture_result["paths"]
        if not image_paths:
            return ToolResult.error_result("vision capture completed but no frames were saved")

        images = [tool_image_from_path(path, "image/jpeg") for path in image_paths]
        output_lines = [
            f"captured camera frames for {params.duration_seconds} seconds.",
            f"saved {len(image_paths)} representative frames for the next visual reasoning step.",
            "frame files:",
        ]
        output_lines.extend(f"- {path}" for path in image_paths)
        return ToolResult.success_result(
            output="\n".join(output_lines),
            images=images,
            metadata={
                "duration_seconds": params.duration_seconds,
                "camera_index": params.camera_index,
                "frame_count": capture_result["frame_count"],
                "image_paths": image_paths,
                "image_mime_types": ["image/jpeg"] * len(image_paths),
                "visual_context_message": (
                    "camera frames from the previous tool call are attached. "
                    "use them as the latest visual scene context before deciding the next action."
                ),
            },
        )

    async def execute(self, invocation: ToolInvocation) -> ToolResult:
        try:
            return await self._execute(invocation)
        except Exception as e:
            return ToolResult.error_result(f"error executing vision capture tool with error: {e}")



from setuptools import setup, find_packages

setup(
    name='pegasus-ai',
    version='0.1.6',
    description='A lightweight terminal agent framework for long-running modular agents',
    url='https://github.com/asuzukosi/pegasus',
    author='Asuzu Kosi',
    author_email='keloasuzu@yahoo.com',
    license='BSD 2-clause',
    packages=find_packages(exclude=['tests*', 'docs*', 'assets*', 'pegasus.core*']),
    include_package_data=True,
    package_data={'pegasus': ['prompts/*.j2']},
    install_requires=[
        'art',
        'click',
        'ddgs',
        'fastmcp',
        'jinja2',
        'openai',
        'opencv-python',
        'pillow',
        'platformdirs',
        'playwright',
        'pydantic',
        'python-dotenv',
        'retry',
        'rich',
        'tiktoken',
        'tomli',
    ],
    entry_points={
        'console_scripts': [
            'pegasus-cli=pegasus.cli.cli:run_cli',
        ],
    },
    python_requires='>=3.9',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
)



from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List
import os

from fastmcp import Client
from fastmcp.client.transports import SSETransport, StdioTransport

from pegasus.config.config import MCPServerConfig


class MCPClientStatus(str, Enum):
    DISCONNECTED = "disconnected"
    CONNECTING = "connecting"
    CONNECTED = "connected"
    ERROR = "error"


@dataclass
class MCPToolInfo:
    name: str
    description: str
    input_schema: dict[str, Any] = field(default_factory=dict)
    server_name: str = ""


class MCPClient:
    def __init__(self, name: str, mcp_config: MCPServerConfig, cwd: Path):
        self.name = name
        self.mcp_config = mcp_config
        self.cwd = cwd
        self.status = MCPClientStatus.DISCONNECTED
        self.client: Client | None = None
        self._tools: Dict[str, MCPToolInfo] = {}

    @property
    def tools(self) -> List[MCPToolInfo]:
        return [value for value in self._tools.values()]

    def _resolve_server_cwd(self) -> Path:
        if self.mcp_config.cwd is None:
            return self.cwd.resolve()
        if self.mcp_config.cwd.is_absolute():
            return self.mcp_config.cwd.resolve()
        return (self.cwd / self.mcp_config.cwd).resolve()

    def _create_transport(self) -> SSETransport | StdioTransport:
        if self.mcp_config.command:
            all_envs = os.environ.copy()
            all_envs.update(self.mcp_config.env)
            cwd = self._resolve_server_cwd().as_posix()
            return StdioTransport(self.mcp_config.command, self.mcp_config.args, all_envs, cwd)
        if self.mcp_config.url:
            return SSETransport(self.mcp_config.url)
        raise ValueError("No transport provided")

    async def connect(self) -> None:
        if self.status == MCPClientStatus.CONNECTED:
            return
        self.status = MCPClientStatus.CONNECTING
        try:
            self.client = Client(transport=self._create_transport())
            await self.client.__aenter__()
            tools = await self.client.list_tools()
            self._tools.clear()
            for tool in tools:
                self._tools[tool.name] = MCPToolInfo(
                    name=tool.name,
                    description=getattr(tool, "description", "") or "",
                    input_schema=getattr(tool, "inputSchema", {}) or {},
                    server_name=self.name,
                )
            self.status = MCPClientStatus.CONNECTED
        except Exception:
            self.status = MCPClientStatus.ERROR
            raise

    async def disconnect(self) -> None:
        if self.client:
            await self.client.__aexit__(None, None, None)
            self.client = None
        self._tools.clear()
        self.status = MCPClientStatus.DISCONNECTED

    async def call_tool(self, tool_name: str, params: dict[str, Any]) -> dict[str, Any]:
        if self.client is None or self.status != MCPClientStatus.CONNECTED:
            raise RuntimeError("mcp server is not connected")
        if tool_name not in self._tools:
            raise ValueError(f"tool {tool_name} not found in mcp server {self.name}")
        result = await self.client.call_tool(tool_name, params)
        output: list[str] = []
        for item in getattr(result, "content", []) or []:
            if hasattr(item, "text") and item.text:
                output.append(item.text)
            elif hasattr(item, "data") and item.data is not None:
                output.append(str(item.data))
            else:
                output.append(str(item))
        return {
            "output": "\n".join(output).strip(),
            "is_error": getattr(result, "is_error", False),
            "raw_content": output,
        }



from __future__ import annotations

import asyncio
import ast
import json
import re
import textwrap
from types import SimpleNamespace
from dataclasses import dataclass
from enum import Enum
from typing import Any

from pydantic import BaseModel, Field

from pegasus.config.config import Config
from pegasus.mcp.client import MCPClient, MCPToolInfo
from pegasus.tools.base import Tool
from pegasus.tools.data import ToolInvocation, ToolResult, ToolType


class MCPExecutorAction(str, Enum):
    LIST = "list"
    EXECUTE = "execute"

    def __str__(self) -> str:
        return self.value


class MCPExecutorParams(BaseModel):
    action: str = Field(
        ...,
        description="The action to perform on the mcp executor. Actions are: list, execute",
    )
    code: str | None = Field(
        None,
        description="Python code to execute when action is 'execute'. Set a variable named result to control the final returned value.",
    )


@dataclass
class MCPFunctionBinding:
    python_name: str
    tool_name: str
    server_name: str
    tool_info: MCPToolInfo


class MCPExecutorTool(Tool):
    name: str = "mcp_executor"
    description: str = (
        "Execute mcp capabilities through a compact python interface. "
        "Use action='list' to inspect available mcp functions and action='execute' "
        "to run python that awaits those functions and sets a result variable."
    )
    type: ToolType = ToolType.MCP
    schema: MCPExecutorParams = MCPExecutorParams

    def __init__(self, config: Config) -> None:
        self._config = config
        self._clients: dict[str, MCPClient] = {}
        self._bindings: dict[str, MCPFunctionBinding] = {}
        self._initialized = False
        self._lock = asyncio.Lock()

    async def _ensure_initialized(self) -> None:
        if self._initialized:
            return
        async with self._lock:
            if self._initialized:
                return
            self._clients = {}
            self._bindings = {}
            mcp_manager = self._config.mcp_manager
            if mcp_manager is None or not mcp_manager.initialized:
                self._initialized = True
                return
            for server_name, client in mcp_manager.connected_clients.items():
                self._clients[server_name] = client
                for tool_info in client.tools:
                    python_name = self._build_python_name(server_name, tool_info.name)
                    self._bindings[python_name] = MCPFunctionBinding(
                        python_name=python_name,
                        tool_name=tool_info.name,
                        server_name=server_name,
                        tool_info=tool_info,
                    )
            self._initialized = True

    def _build_python_name(self, server_name: str, tool_name: str) -> str:
        candidate = f"{server_name}__{tool_name}"
        candidate = re.sub(r"[^0-9a-zA-Z_]", "_", candidate)
        candidate = re.sub(r"_+", "_", candidate).strip("_")
        if not candidate:
            candidate = "mcp_function"
        if candidate[0].isdigit():
            candidate = f"mcp_{candidate}"
        return candidate.lower()

    def _format_schema(self, schema: dict[str, Any]) -> str:
        properties = schema.get("properties", {}) if isinstance(schema, dict) else {}
        required = set(schema.get("required", [])) if isinstance(schema, dict) else set()
        if not properties:
            return "{}"
        lines = ["{"]
        for key, value in properties.items():
            type_name = value.get("type", "any") if isinstance(value, dict) else "any"
            suffix = " required" if key in required else ""
            lines.append(f'  "{key}": "{type_name}{suffix}"')
        lines.append("}")
        return "\n".join(lines)

    def _render_function_list(self) -> str:
        if not self._bindings:
            return "no connected mcp servers are available."
        lines = [
            "available mcp python functions:",
            "",
            "call them with await inside execute mode and set a result variable.",
            "a limited asyncio helper is already available, so use asyncio.gather(...) without importing asyncio.",
            "",
        ]
        for binding in sorted(self._bindings.values(), key=lambda item: item.python_name):
            description = binding.tool_info.description or "no description provided"
            lines.append(f"- python name: {binding.python_name}")
            lines.append(f"  server: {binding.server_name}")
            lines.append(f"  mcp tool: {binding.tool_name}")
            lines.append(f"  description: {description}")
            lines.append("  arguments:")
            lines.append(textwrap.indent(self._format_schema(binding.tool_info.input_schema), "    "))
            lines.append("")
        return "\n".join(lines).rstrip()

    def _validate_script(self, code: str) -> list[str]:
        blocked_calls = {
            "open",
            "exec",
            "eval",
            "compile",
            "input",
            "__import__",
            "globals",
            "locals",
            "vars",
            "getattr",
            "setattr",
            "delattr",
            "breakpoint",
        }
        blocked_nodes = (
            ast.Import,
            ast.ImportFrom,
            ast.FunctionDef,
            ast.AsyncFunctionDef,
            ast.ClassDef,
            ast.With,
            ast.AsyncWith,
            ast.Try,
            ast.Raise,
            ast.Delete,
            ast.Global,
            ast.Nonlocal,
            ast.Lambda,
        )
        errors: list[str] = []
        try:
            tree = ast.parse(code, mode="exec")
        except SyntaxError as e:
            return [f"syntax error: {e}"]

        for node in ast.walk(tree):
            if isinstance(node, blocked_nodes):
                errors.append(f"unsupported syntax: {type(node).__name__}")
            if isinstance(node, ast.Name) and node.id.startswith("_"):
                errors.append(f"unsafe name access: {node.id}")
            if isinstance(node, ast.Attribute) and node.attr.startswith("_"):
                errors.append(f"unsafe attribute access: {node.attr}")
            if isinstance(node, ast.Call):
                func = node.func
                if isinstance(func, ast.Name) and func.id in blocked_calls:
                    errors.append(f"blocked call: {func.id}")
        return sorted(set(errors))

    def _safe_builtins(self, print_buffer: list[str]) -> dict[str, Any]:
        def sandbox_print(*args, **kwargs) -> None:
            sep = kwargs.get("sep", " ")
            end = kwargs.get("end", "\n")
            print_buffer.append(sep.join(str(arg) for arg in args) + end.rstrip("\n"))

        return {
            "print": sandbox_print,
            "len": len,
            "str": str,
            "int": int,
            "float": float,
            "bool": bool,
            "dict": dict,
            "list": list,
            "tuple": tuple,
            "set": set,
            "min": min,
            "max": max,
            "sum": sum,
            "enumerate": enumerate,
            "range": range,
            "zip": zip,
            "sorted": sorted,
            "any": any,
            "all": all,
        }

    def _coerce_binding_params(
        self,
        binding: MCPFunctionBinding,
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> dict[str, Any]:
        if not args:
            return kwargs
        schema = binding.tool_info.input_schema if isinstance(binding.tool_info.input_schema, dict) else {}
        properties = schema.get("properties", {}) if isinstance(schema, dict) else {}
        property_names = list(properties.keys())
        if len(args) > len(property_names):
            raise ValueError(
                f"{binding.python_name} accepts at most {len(property_names)} positional arguments, got {len(args)}"
            )
        params = dict(kwargs)
        for name, value in zip(property_names, args, strict=False):
            if name in params:
                raise ValueError(f"{binding.python_name} received multiple values for argument '{name}'")
            params[name] = value
        return params

    async def _run_binding(self, binding: MCPFunctionBinding, *args: Any, **kwargs: Any) -> dict[str, Any]:
        params = self._coerce_binding_params(binding, args, kwargs)
        client = self._clients[binding.server_name]
        result = await client.call_tool(binding.tool_name, params)
        if result["is_error"]:
            raise RuntimeError(result["output"] or f"mcp tool {binding.tool_name} failed")
        return {
            "server": binding.server_name,
            "tool": binding.tool_name,
            "output": result["output"],
            "raw_content": result["raw_content"],
        }

    async def _execute_script(self, code: str) -> ToolResult:
        validation_errors = self._validate_script(code)
        if validation_errors:
            return ToolResult.error_result(
                "unsafe or unsupported python provided for mcp execution",
                output="\n".join(validation_errors),
            )

        print_buffer: list[str] = []
        namespace: dict[str, Any] = {
            "__builtins__": self._safe_builtins(print_buffer),
            "asyncio": SimpleNamespace(gather=asyncio.gather, sleep=asyncio.sleep),
            "json": json,
        }
        for python_name, binding in self._bindings.items():
            async def _wrapper(*args, _binding: MCPFunctionBinding = binding, **kwargs):
                return await self._run_binding(_binding, *args, **kwargs)
            namespace[python_name] = _wrapper

        function_source = (
            "async def __mcp_user_code__():\n"
            + textwrap.indent(code, "    ")
            + "\n    try:\n"
            + "        return result\n"
            + "    except NameError:\n"
            + "        return None\n"
        )
        exec(function_source, namespace, namespace)
        result_value = await namespace["__mcp_user_code__"]()

        sections: list[str] = []
        if print_buffer:
            sections.append("printed output:\n" + "\n".join(line for line in print_buffer if line))
        if result_value is not None:
            try:
                rendered_result = json.dumps(result_value, indent=2, default=str)
            except TypeError:
                rendered_result = str(result_value)
            sections.append("result:\n" + rendered_result)
        if not sections:
            sections.append("script completed without printed output or result.")

        return ToolResult.success_result(
            output="\n\n".join(sections),
            metadata={
                "action": "execute",
                "available_functions": sorted(self._bindings.keys()),
            },
        )

    async def _execute(self, invocation: ToolInvocation) -> ToolResult:
        params = MCPExecutorParams(**invocation.params)
        await self._ensure_initialized()

        if params.action == MCPExecutorAction.LIST:
            return ToolResult.success_result(
                output=self._render_function_list(),
                metadata={"action": "list", "function_count": len(self._bindings)},
            )
        if params.action == MCPExecutorAction.EXECUTE:
            if not params.code or not params.code.strip():
                return ToolResult.error_result("code is required when action is 'execute'")
            if not self._bindings:
                return ToolResult.error_result("no mcp functions are available because no mcp servers are connected")
            return await self._execute_script(params.code)
        return ToolResult.error_result(
            f"invalid action: {params.action}. valid actions are: {', '.join(action.value for action in MCPExecutorAction)}"
        )

    async def execute(self, invocation: ToolInvocation) -> ToolResult:
        try:
            return await self._execute(invocation)
        except Exception as e:
            return ToolResult.error_result(f"error executing mcp executor with error: {e}")

    async def close(self) -> None:
        self._clients = {}
        self._bindings = {}
        self._initialized = False
